##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
OpenDirectory service tests.
"""

from uuid import UUID

from twisted.trial import unittest

from ...expression import (
    CompoundExpression, Operand, MatchExpression, MatchType, MatchFlags
)
from .._constants import ODAttribute
from .._service import DirectoryService
from ...idirectory import QueryNotSupportedError


class OpenDirectoryServiceTestCase(unittest.TestCase):
    """
    Tests for L{DirectoryService}.
    """

    def setUp(self):
        self.service = DirectoryService()

    def tearDown(self):
        self.service._deletePool()


    def test_queryFromMatchExpression_recordType(self):
        """
        Make sure queryFromMatchExpression handles recordType correctly
        """
        query = self.service._queryFromMatchExpression(
            MatchExpression(
                self.service.fieldName.shortNames, u"xyzzy"
            ),
            recordType=self.service.recordType.group
        )
        # FIXME:
        # Actually, how do we inspect the query object to peek at the
        # recordType that got set?
        query

    test_queryFromMatchExpression_recordType.todo = ""


    def test_queryStringFromMatchExpression_matchTypes(self):
        """
        Match expressions with each match type produces the correct
        operator=value string.
        """
        for matchType, expected in (
            (MatchType.equals, u"=xyzzy"),
            (MatchType.startsWith, u"=xyzzy*"),
            (MatchType.endsWith, u"=*xyzzy"),
            (MatchType.contains, u"=*xyzzy*"),
            (MatchType.lessThan, u"<xyzzy"),
            (MatchType.greaterThan, u">xyzzy"),
            (MatchType.lessThanOrEqualTo, u"<=xyzzy"),
            (MatchType.greaterThanOrEqualTo, u">=xyzzy"),
        ):
            expression = MatchExpression(
                self.service.fieldName.shortNames, u"xyzzy",
                matchType=matchType
            )
            queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
            self.assertEquals(
                recordTypes,
                set(
                    [
                        u"dsRecTypeStandard:Users",
                        u"dsRecTypeStandard:Groups",
                        u"dsRecTypeStandard:Places",
                        u"dsRecTypeStandard:Resources",
                    ]
                )
            )
            self.assertEquals(
                queryString,
                u"({attribute}{expected})".format(
                    attribute=ODAttribute.shortName.value, expected=expected
                )
            )


    def test_queryStringFromMatchExpression_match_not(self):
        """
        Match expression with the C{NOT} flag adds the C{!} operator.
        """
        expression = MatchExpression(
            self.service.fieldName.shortNames, u"xyzzy",
            flags=MatchFlags.NOT
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(
            recordTypes,
            set(
                [
                    u"dsRecTypeStandard:Users",
                    u"dsRecTypeStandard:Groups",
                    u"dsRecTypeStandard:Places",
                    u"dsRecTypeStandard:Resources",
                ]
            )
        )
        self.assertEquals(
            queryString,
            u"(!{attribute}=xyzzy)".format(
                attribute=ODAttribute.shortName.value,
            )
        )


    def test_queryStringFromMatchExpression_match_caseInsensitive(self):
        """
        Match expression with the C{caseInsensitive} flag adds the C{??????}
        operator.
        """
        expression = MatchExpression(
            self.service.fieldName.shortNames, u"xyzzy",
            flags=MatchFlags.caseInsensitive
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(
            recordTypes,
            set(
                [
                    u"dsRecTypeStandard:Users",
                    u"dsRecTypeStandard:Groups",
                    u"dsRecTypeStandard:Places",
                    u"dsRecTypeStandard:Resources",
                ]
            )
        )
        self.assertEquals(
            queryString,
            u"???????({attribute}=xyzzy)".format(
                attribute=ODAttribute.shortName.value,
            )
        )

    test_queryStringFromMatchExpression_match_caseInsensitive.todo = (
        "unimplemented"
    )


    def test_queryStringFromMatchExpression_match_quoting(self):
        """
        Special characters are quoted properly.
        """
        expression = MatchExpression(
            self.service.fieldName.fullNames,
            u"\\xyzzy: a/b/(c)* ~~ >=< ~~ &| \0!!"
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(
            recordTypes,
            set(
                [
                    u"dsRecTypeStandard:Groups", u"dsRecTypeStandard:Users",
                    u"dsRecTypeStandard:Places", u"dsRecTypeStandard:Resources",
                ]
            )
        )
        self.assertEquals(
            queryString,
            u"({attribute}={expected})".format(
                attribute=ODAttribute.fullName.value,
                expected=(
                    u"\\5Cxyzzy: a\\2Fb\\2F\\28c\\29\\2A "
                    "\\7E\\7E \\3E\\3D\\3C \\7E\\7E \\26\\7C \\00!!"
                )
            )
        )


    def test_queryStringFromExpression(self):
        # CompoundExpressions

        expression = CompoundExpression(
            [
                MatchExpression(
                    self.service.fieldName.uid, u"a",
                    matchType=MatchType.contains
                ),
                MatchExpression(
                    self.service.fieldName.guid, UUID(int=0),
                    matchType=MatchType.contains
                ),
                MatchExpression(
                    self.service.fieldName.shortNames, u"c",
                    matchType=MatchType.contains
                ),
                MatchExpression(
                    self.service.fieldName.emailAddresses, u"d",
                    matchType=MatchType.startsWith
                ),
                MatchExpression(
                    self.service.fieldName.fullNames, u"e",
                    matchType=MatchType.equals
                ),
            ],
            Operand.AND
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(
            recordTypes,
            set(
                [
                    u"dsRecTypeStandard:Users",
                    u"dsRecTypeStandard:Groups",
                    u"dsRecTypeStandard:Places",
                    u"dsRecTypeStandard:Resources",
                ]
            )
        )
        self.assertEquals(
            queryString,
            (
                u"(&(dsAttrTypeStandard:GeneratedUID=*a*)"
                u"(dsAttrTypeStandard:GeneratedUID="
                u"*00000000-0000-0000-0000-000000000000*)"
                u"(dsAttrTypeStandard:RecordName=*c*)"
                u"(dsAttrTypeStandard:EMailAddress=d*)"
                u"(dsAttrTypeStandard:RealName=e))"
            )
        )

        expression = CompoundExpression(
            [
                MatchExpression(
                    self.service.fieldName.shortNames, u"a",
                    matchType=MatchType.contains
                ),
                MatchExpression(
                    self.service.fieldName.emailAddresses, u"b",
                    matchType=MatchType.startsWith
                ),
                MatchExpression(
                    self.service.fieldName.fullNames, u"c",
                    matchType=MatchType.equals
                ),
            ],
            Operand.OR
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(
            recordTypes,
            set(
                [
                    u"dsRecTypeStandard:Users",
                    u"dsRecTypeStandard:Groups",
                    u"dsRecTypeStandard:Places",
                    u"dsRecTypeStandard:Resources",
                ]
            )
        )
        self.assertEquals(
            queryString,
            (
                u"(|(dsAttrTypeStandard:RecordName=*a*)"
                u"(dsAttrTypeStandard:EMailAddress=b*)"
                u"(dsAttrTypeStandard:RealName=c))"
            )
        )


    def test_queryStringFromExpression_recordType(self):
        """
        Record type in expression
        """
        # AND expression
        expression = CompoundExpression(
            [
                MatchExpression(
                    self.service.fieldName.shortNames,
                    u"xyzzy",
                    matchType=MatchType.equals
                ),
                MatchExpression(
                    self.service.fieldName.recordType,
                    self.service.recordType.group,
                    matchType=MatchType.equals
                ),
                MatchExpression(
                    self.service.fieldName.recordType,
                    self.service.recordType.user,
                    matchType=MatchType.equals
                ),
            ],
            Operand.AND
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(recordTypes, set())
        self.assertEquals(
            queryString,
            u"(dsAttrTypeStandard:RecordName=xyzzy)"
        )

        # AND subexpression
        expression = CompoundExpression(
            [
                MatchExpression(
                    self.service.fieldName.shortNames,
                    u"xxxxx",
                    matchType=MatchType.equals
                ),
                CompoundExpression(
                    [

                        MatchExpression(
                            self.service.fieldName.recordType,
                            self.service.recordType.group,
                            matchType=MatchType.equals
                        ),
                        MatchExpression(
                            self.service.fieldName.shortNames,
                            u"yyyyy",
                            matchType=MatchType.equals
                        ),
                    ],
                    Operand.AND
                ),
            ],
            Operand.OR
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(set(recordTypes), set([u"dsRecTypeStandard:Groups", ]))
        self.assertEquals(
            queryString,
            u"("
            u"|(dsAttrTypeStandard:RecordName=xxxxx)"
            u"(dsAttrTypeStandard:RecordName=yyyyy)"
            u")"
        )
        # NOR expression
        expression = CompoundExpression(
            [
                MatchExpression(
                    self.service.fieldName.shortNames,
                    u"xxxxx",
                    matchType=MatchType.equals
                ),
                MatchExpression(
                    self.service.fieldName.shortNames,
                    u"yyyyy",
                    matchType=MatchType.equals
                ),
                MatchExpression(
                    self.service.fieldName.recordType,
                    self.service.recordType.user,
                    matchType=MatchType.equals,
                    flags=MatchFlags.NOT
                ),
            ],
            Operand.OR
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(
            set(recordTypes),
            set(
                [
                    u"dsRecTypeStandard:Groups",
                    u"dsRecTypeStandard:Places",
                    u"dsRecTypeStandard:Resources",
                ]
            )
        )
        self.assertEquals(
            queryString,
            u"("
            u"|(dsAttrTypeStandard:RecordName=xxxxx)"
            u"(dsAttrTypeStandard:RecordName=yyyyy)"
            u")"
        )

        # Simple AND expression -> empty query string
        expression = CompoundExpression(
            [
                MatchExpression(
                    self.service.fieldName.recordType,
                    self.service.recordType.user,
                    matchType=MatchType.equals
                ),
            ],
            Operand.AND
        )
        queryString, recordTypes = self.service._queryStringAndRecordTypesFromExpression(expression)
        self.assertEquals(set(recordTypes), set([u"dsRecTypeStandard:Users"]))
        self.assertEquals(
            queryString,
            u""
        )

        # recordType OR expression raises QueryNotSupportedError
        expression = CompoundExpression(
            [
                MatchExpression(
                    self.service.fieldName.recordType,
                    self.service.recordType.user,
                    matchType=MatchType.equals
                ),
            ],
            Operand.OR
        )
        self.assertRaises(
            QueryNotSupportedError,
            self.service._queryStringAndRecordTypesFromExpression,
            expression,
        )



##
# Copyright (c) 2005-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Extentions to twisted.internet.tcp.
"""

__all__ = [
    "MaxAcceptTCPServer",
    "MaxAcceptSSLServer",
]

import socket
from OpenSSL import SSL

from twisted.application import internet
from twisted.internet import tcp, ssl
from twisted.internet.defer import succeed

from twext.python.log import Logger

log = Logger()



class MaxAcceptPortMixin(object):
    """
    Mixin for resetting maxAccepts.
    """
    def doRead(self):
        self.numberAccepts = min(
            self.factory.maxRequests - self.factory.outstandingRequests,
            self.factory.maxAccepts
        )
        tcp.Port.doRead(self)



class MaxAcceptTCPPort(MaxAcceptPortMixin, tcp.Port):
    """
    Use for non-inheriting tcp ports.
    """



class MaxAcceptSSLPort(MaxAcceptPortMixin, ssl.Port):
    """
    Use for non-inheriting SSL ports.
    """



class InheritedTCPPort(MaxAcceptTCPPort):
    """
    A tcp port which uses an inherited file descriptor.
    """

    def __init__(self, fd, factory, reactor):
        tcp.Port.__init__(self, 0, factory, reactor=reactor)
        # MOR: careful because fromfd dup()'s the socket, so we need to
        # make sure we don't leak file descriptors
        self.socket = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
        self._realPortNumber = self.port = self.socket.getsockname()[1]


    def createInternetSocket(self):
        return self.socket


    def startListening(self):
        log.info(
            "{self.factory.__class__} starting on {self._realPortNumber}",
            self=self
        )
        self.factory.doStart()
        self.connected = 1
        self.fileno = self.socket.fileno
        self.numberAccepts = self.factory.maxRequests
        self.startReading()



class InheritedSSLPort(InheritedTCPPort):
    """
    An SSL port which uses an inherited file descriptor.
    """

    _socketShutdownMethod = 'sock_shutdown'

    transport = ssl.Server


    def __init__(self, fd, factory, ctxFactory, reactor):
        InheritedTCPPort.__init__(self, fd, factory, reactor)
        self.ctxFactory = ctxFactory
        self.socket = SSL.Connection(self.ctxFactory.getContext(), self.socket)


    def _preMakeConnection(self, transport):
        transport._startTLS()
        return tcp.Port._preMakeConnection(self, transport)



def _allConnectionsClosed(protocolFactory):
    """
    Check to see if protocolFactory implements allConnectionsClosed( ) and
    if so, call it.  Otherwise, return immediately.
    This allows graceful shutdown by waiting for all requests to be completed.

    @param protocolFactory: (usually) an HTTPFactory implementing
        allConnectionsClosed which returns a Deferred which fires when all
        connections are closed.

    @return: A Deferred firing None when all connections are closed, or
        immediately if the given factory does not track its connections (e.g.
        InheritingProtocolFactory)
    """
    if hasattr(protocolFactory, "allConnectionsClosed"):
        return protocolFactory.allConnectionsClosed()
    return succeed(None)



class MaxAcceptTCPServer(internet.TCPServer):
    """
    TCP server which will uses MaxAcceptTCPPorts (and optionally,
    inherited ports)

    @ivar myPort: When running, this is set to the L{IListeningPort} being
        managed by this service.
    """

    def __init__(self, *args, **kwargs):
        internet.TCPServer.__init__(self, *args, **kwargs)
        self.protocolFactory = self.args[1]
        self.protocolFactory.myServer = self
        self.inherit = self.kwargs.get("inherit", False)
        self.backlog = self.kwargs.get("backlog", None)
        self.interface = self.kwargs.get("interface", None)


    def _getPort(self):
        from twisted.internet import reactor

        if self.inherit:
            port = InheritedTCPPort(self.args[0], self.args[1], reactor)
        else:
            port = MaxAcceptTCPPort(
                self.args[0], self.args[1],
                self.backlog, self.interface, reactor
            )

        port.startListening()
        self.myPort = port
        return port


    def stopService(self):
        """
        Wait for outstanding requests to finish

        @return: a Deferred which fires when all outstanding requests are
            complete
        """
        internet.TCPServer.stopService(self)
        return _allConnectionsClosed(self.protocolFactory)



class MaxAcceptSSLServer(internet.SSLServer):
    """
    SSL server which will uses MaxAcceptSSLPorts (and optionally,
    inherited ports)
    """

    def __init__(self, *args, **kwargs):
        internet.SSLServer.__init__(self, *args, **kwargs)
        self.protocolFactory = self.args[1]
        self.protocolFactory.myServer = self
        self.inherit = self.kwargs.get("inherit", False)
        self.backlog = self.kwargs.get("backlog", None)
        self.interface = self.kwargs.get("interface", None)


    def _getPort(self):
        from twisted.internet import reactor

        if self.inherit:
            port = InheritedSSLPort(
                self.args[0], self.args[1], self.args[2], reactor
            )
        else:
            port = MaxAcceptSSLPort(
                self.args[0], self.args[1], self.args[2],
                self.backlog, self.interface, self.reactor
            )

        port.startListening()
        self.myPort = port
        return port


    def stopService(self):
        """
        Wait for outstanding requests to finish

        @return: a Deferred which fires when all outstanding requests are
            complete.
        """
        internet.SSLServer.stopService(self)
        # TODO: check for an ICompletionWaiter interface
        return _allConnectionsClosed(self.protocolFactory)



# -*- test-case-name: twext.python.test.test_parallel -*-
##
# Copyright (c) 2012-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Utilities for parallelizing tasks.
"""

from twisted.internet.defer import inlineCallbacks, DeferredList, returnValue

class Parallelizer(object):
    """
    Do some operation with a degree of parallelism, using a set of resources
    which may each only be used for one task at a time, given some underlying
    API that returns L{Deferreds}.

    @ivar available: A list of available resources from the C{resources}
        constructor parameter.

    @ivar busy: A list of resources which are currently being used by
        operations.
    """

    def __init__(self, resources):
        """
        Initialize a L{Parallelizer} with a list of objects that will be passed
        to the callables sent to L{Parallelizer.do}.

        @param resources: objects which may be of any arbitrary type.
        @type resources: C{list}
        """
        self.available = list(resources)
        self.busy = []
        self.activeDeferreds = []


    @inlineCallbacks
    def do(self, operation):
        """
        Call C{operation} with one of the resources in C{self.available},
        removing that value for use by other callers of C{do} until the task
        performed by C{operation} is complete (in other words, the L{Deferred}
        returned by C{operation} has fired).

        @param operation: a 1-argument callable taking a resource from
            C{self.active} and returning a L{Deferred} when it's done using
            that resource.
        @type operation: C{callable}

        @return: a L{Deferred} that fires as soon as there are resources
            available such that this task can be I{started} - not completed.
        """
        if not self.available:
            yield DeferredList(self.activeDeferreds, fireOnOneCallback=True,
                               fireOnOneErrback=True)
        active = self.available.pop(0)
        self.busy.append(active)
        o = operation(active)
        def andFinally(whatever):
            self.activeDeferreds.remove(o)
            self.busy.remove(active)
            self.available.append(active)
            return whatever
        self.activeDeferreds.append(o)
        o.addBoth(andFinally)
        returnValue(None)


    def done(self):
        """
        Wait until all operations started by L{Parallelizer.do} are completed.

        @return: a L{Deferred} that fires (with C{None}) when all the currently
            pending work on this L{Parallelizer} is completed and C{busy} is
            empty again.
        """
        return (DeferredList(self.activeDeferreds)
                .addCallback(lambda ignored: None))



##
# Copyright (c) 2012-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Test cases for L{twext.enterprise.dal.record}.
"""

import datetime

from twisted.internet.defer import inlineCallbacks, gatherResults, returnValue
from twisted.trial.unittest import TestCase, SkipTest

from twext.enterprise.dal.record import (
    Record, fromTable, ReadOnly, NoSuchRecord,
    SerializableRecord)
from twext.enterprise.dal.test.test_parseschema import SchemaTestHelper
from twext.enterprise.dal.syntax import SchemaSyntax
from twext.enterprise.fixtures import buildConnectionPool

# from twext.enterprise.dal.syntax import


sth = SchemaTestHelper()
sth.id = lambda: __name__
schemaString = """
create table ALPHA (BETA integer primary key, GAMMA text);
create table DELTA (PHI integer primary key default (nextval('myseq')),
                    EPSILON text not null,
                    ZETA timestamp not null default '2012-12-12 12:12:12' );
"""

# sqlite can be made to support nextval() as a function, but 'create sequence'
# is syntax and can't.
parseableSchemaString = """
create sequence myseq;
""" + schemaString

try:
    testSchema = SchemaSyntax(sth.schemaFromString(parseableSchemaString))
except SkipTest as e:
    Alpha = Delta = object
    skip = e
else:
    Alpha = fromTable(testSchema.ALPHA)
    Delta = fromTable(testSchema.DELTA)
    skip = False



class TestRecord(Record, Alpha):
    """
    A sample test record.
    """



class TestSerializeRecord(SerializableRecord, Alpha):
    """
    A sample test serializable record with default values specified.
    """



class TestAutoRecord(Record, Delta):
    """
    A sample test record with default values specified.
    """



class TestCRUD(TestCase):
    """
    Tests for creation, mutation, and deletion operations.
    """

    def setUp(self):
        self.pool = buildConnectionPool(self, schemaString)


    @inlineCallbacks
    def test_simpleLoad(self):
        """
        Loading an existing row from the database by its primary key will
        populate its attributes from columns of the corresponding row in the
        database.
        """
        txn = self.pool.connection()
        yield txn.execSQL("insert into ALPHA values (:1, :2)", [234, "one"])
        yield txn.execSQL("insert into ALPHA values (:1, :2)", [456, "two"])

        rec = yield TestRecord.load(txn, 456)
        self.assertIsInstance(rec, TestRecord)
        self.assertEquals(rec.beta, 456)
        self.assertEquals(rec.gamma, "two")

        rec2 = yield TestRecord.load(txn, 234)
        self.assertIsInstance(rec2, TestRecord)
        self.assertEqual(rec2.beta, 234)
        self.assertEqual(rec2.gamma, "one")


    @inlineCallbacks
    def test_missingLoad(self):
        """
        Try loading an row which doesn't exist
        """
        txn = self.pool.connection()
        yield txn.execSQL("insert into ALPHA values (:1, :2)", [234, "one"])
        yield self.assertFailure(TestRecord.load(txn, 456), NoSuchRecord)


    @inlineCallbacks
    def test_simpleCreate(self):
        """
        When a record object is created, a row with matching column values will
        be created in the database.
        """
        txn = self.pool.connection()

        rec = yield TestRecord.create(txn, beta=3, gamma=u'epsilon')
        self.assertEquals(rec.beta, 3)
        self.assertEqual(rec.gamma, u'epsilon')

        rows = yield txn.execSQL("select BETA, GAMMA from ALPHA")
        self.assertEqual(rows, [tuple([3, u'epsilon'])])


    @inlineCallbacks
    def test_simpleDelete(self):
        """
        When a record object is deleted, a row with a matching primary key will
        be deleted in the database.
        """
        txn = self.pool.connection()

        def mkrow(beta, gamma):
            return txn.execSQL("insert into ALPHA values (:1, :2)",
                               [beta, gamma])

        yield gatherResults(
            [mkrow(123, u"one"), mkrow(234, u"two"), mkrow(345, u"three")]
        )
        tr = yield TestRecord.load(txn, 234)
        yield tr.delete()
        rows = yield txn.execSQL("select BETA, GAMMA from ALPHA order by BETA")
        self.assertEqual(rows, [(123, u"one"), (345, u"three")])


    @inlineCallbacks
    def oneRowCommitted(self, beta=123, gamma=u'456'):
        """
        Create, commit, and return one L{TestRecord}.
        """
        txn = self.pool.connection(self.id())
        row = yield TestRecord.create(txn, beta=beta, gamma=gamma)
        yield txn.commit()
        returnValue(row)


    @inlineCallbacks
    def test_deleteWhenDeleted(self):
        """
        When a record object is deleted, if it's already been deleted, it will
        raise L{NoSuchRecord}.
        """
        row = yield self.oneRowCommitted()
        txn = self.pool.connection(self.id())
        newRow = yield TestRecord.load(txn, row.beta)
        yield newRow.delete()
        yield self.assertFailure(newRow.delete(), NoSuchRecord)


    @inlineCallbacks
    def test_cantCreateWithoutRequiredValues(self):
        """
        When a L{Record} object is created without required values, it raises a
        L{TypeError}.
        """
        txn = self.pool.connection()
        te = yield self.assertFailure(TestAutoRecord.create(txn), TypeError)
        self.assertIn("required attribute 'epsilon' not passed", str(te))


    @inlineCallbacks
    def test_datetimeType(self):
        """
        When a L{Record} references a timestamp column, it retrieves the date
        as UTC.
        """
        txn = self.pool.connection()
        # Create ...
        rec = yield TestAutoRecord.create(txn, epsilon=1)
        self.assertEquals(
            rec.zeta,
            datetime.datetime(2012, 12, 12, 12, 12, 12)
        )
        yield txn.commit()
        # ... should have the same effect as loading.

        txn = self.pool.connection()
        rec = (yield TestAutoRecord.all(txn))[0]
        self.assertEquals(
            rec.zeta,
            datetime.datetime(2012, 12, 12, 12, 12, 12)
        )


    @inlineCallbacks
    def test_tooManyAttributes(self):
        """
        When a L{Record} object is created with unknown attributes (those which
        don't map to any column), it raises a L{TypeError}.
        """
        txn = self.pool.connection()
        te = yield self.assertFailure(
            TestRecord.create(
                txn, beta=3, gamma=u'three',
                extraBonusAttribute=u'nope',
                otherBonusAttribute=4321,
            ),
            TypeError
        )
        self.assertIn("extraBonusAttribute, otherBonusAttribute", str(te))


    @inlineCallbacks
    def test_createFillsInPKey(self):
        """
        If L{Record.create} is called without an auto-generated primary key
        value for its row, that value will be generated and set on the returned
        object.
        """
        txn = self.pool.connection()
        tr = yield TestAutoRecord.create(txn, epsilon=u'specified')
        tr2 = yield TestAutoRecord.create(txn, epsilon=u'also specified')
        self.assertEquals(tr.phi, 1)
        self.assertEquals(tr2.phi, 2)


    @inlineCallbacks
    def test_attributesArentMutableYet(self):
        """
        Changing attributes on a database object is not supported yet, because
        it's not entirely clear when to flush the SQL to the database.
        Instead, for the time being, use C{.update}.  When you attempt to set
        an attribute, an error will be raised informing you of this fact, so
        that the error is clear.
        """
        txn = self.pool.connection()
        rec = yield TestRecord.create(txn, beta=7, gamma=u'what')

        def setit():
            rec.beta = 12

        ro = self.assertRaises(ReadOnly, setit)
        self.assertEqual(rec.beta, 7)
        self.assertIn("SQL-backed attribute 'TestRecord.beta' is read-only. "
                      "Use '.update(...)' to modify attributes.", str(ro))


    @inlineCallbacks
    def test_simpleUpdate(self):
        """
        L{Record.update} will change the values on the record and in te
        database.
        """
        txn = self.pool.connection()
        rec = yield TestRecord.create(txn, beta=3, gamma=u'epsilon')
        yield rec.update(gamma=u'otherwise')
        self.assertEqual(rec.gamma, u'otherwise')
        yield txn.commit()
        # Make sure that it persists.
        txn = self.pool.connection()
        rec = yield TestRecord.load(txn, 3)
        self.assertEqual(rec.gamma, u'otherwise')


    @inlineCallbacks
    def test_simpleQuery(self):
        """
        L{Record.query} will allow you to query for a record by its class
        attributes as columns.
        """
        txn = self.pool.connection()
        for beta, gamma in [(123, u"one"), (234, u"two"), (345, u"three"),
                            (356, u"three"), (456, u"four")]:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])
        records = yield TestRecord.query(txn, TestRecord.gamma == u"three")
        self.assertEqual(len(records), 2)
        records.sort(key=lambda x: x.beta)
        self.assertEqual(records[0].beta, 345)
        self.assertEqual(records[1].beta, 356)


    @inlineCallbacks
    def test_querySimple(self):
        """
        L{Record.querysimple} will allow you to query for a record by its class
        attributes as columns.
        """
        txn = self.pool.connection()
        for beta, gamma in [(123, u"one"), (234, u"two"), (345, u"three"),
                            (356, u"three"), (456, u"four")]:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])
        records = yield TestRecord.querysimple(txn, gamma=u"three")
        self.assertEqual(len(records), 2)
        records.sort(key=lambda x: x.beta)
        self.assertEqual(records[0].beta, 345)
        self.assertEqual(records[1].beta, 356)


    @inlineCallbacks
    def test_eq(self):
        """
        L{Record.__eq__} works.
        """
        txn = self.pool.connection()
        data = [(123, u"one"), (456, u"four"), (345, u"three"),
                (234, u"two"), (356, u"three")]
        for beta, gamma in data:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])

        one = yield TestRecord.load(txn, 123)
        one_copy = yield TestRecord.load(txn, 123)
        two = yield TestRecord.load(txn, 234)

        self.assertTrue(one == one_copy)
        self.assertFalse(one == two)


    @inlineCallbacks
    def test_all(self):
        """
        L{Record.all} will return all instances of the record, sorted by
        primary key.
        """
        txn = self.pool.connection()
        data = [(123, u"one"), (456, u"four"), (345, u"three"),
                (234, u"two"), (356, u"three")]
        for beta, gamma in data:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])
        self.assertEqual(
            [(x.beta, x.gamma) for x in (yield TestRecord.all(txn))],
            sorted(data)
        )


    @inlineCallbacks
    def test_updatesome(self):
        """
        L{Record.updatesome} will update all instances of the matching records.
        """
        txn = self.pool.connection()
        data = [(123, u"one"), (456, u"four"), (345, u"three"),
                (234, u"two"), (356, u"three")]
        for beta, gamma in data:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])

        yield TestRecord.updatesome(txn, where=(TestRecord.beta == 123), gamma=u"changed")
        yield txn.commit()

        txn = self.pool.connection()
        records = yield TestRecord.all(txn)
        self.assertEqual(
            set([(record.beta, record.gamma,) for record in records]),
            set([
                (123, u"changed"), (456, u"four"), (345, u"three"),
                (234, u"two"), (356, u"three")
            ])
        )

        yield TestRecord.updatesome(txn, where=(TestRecord.beta.In((234, 345,))), gamma=u"changed-2")
        yield txn.commit()

        txn = self.pool.connection()
        records = yield TestRecord.all(txn)
        self.assertEqual(
            set([(record.beta, record.gamma,) for record in records]),
            set([
                (123, u"changed"), (456, u"four"), (345, u"changed-2"),
                (234, u"changed-2"), (356, u"three")
            ])
        )


    @inlineCallbacks
    def test_deleteall(self):
        """
        L{Record.deleteall} will delete all instances of the record.
        """
        txn = self.pool.connection()
        data = [(123, u"one"), (456, u"four"), (345, u"three"),
                (234, u"two"), (356, u"three")]
        for beta, gamma in data:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])

        yield TestRecord.deleteall(txn)
        all = yield TestRecord.all(txn)
        self.assertEqual(len(all), 0)


    @inlineCallbacks
    def test_deletesome(self):
        """
        L{Record.deletesome} will delete all instances of the matching records.
        """
        txn = self.pool.connection()
        data = [(123, u"one"), (456, u"four"), (345, u"three"),
                (234, u"two"), (356, u"three")]
        for beta, gamma in data:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])

        yield TestRecord.deletesome(txn, TestRecord.gamma == u"three")
        all = yield TestRecord.all(txn)
        self.assertEqual(set([record.beta for record in all]), set((123, 456, 234,)))

        yield TestRecord.deletesome(txn, (TestRecord.gamma == u"one").Or(TestRecord.gamma == u"two"))
        all = yield TestRecord.all(txn)
        self.assertEqual(set([record.beta for record in all]), set((456,)))


    @inlineCallbacks
    def test_deletesimple(self):
        """
        L{Record.deletesimple} will delete all instances of the matching records.
        """
        txn = self.pool.connection()
        data = [(123, u"one"), (456, u"four"), (345, u"three"),
                (234, u"two"), (356, u"three")]
        for beta, gamma in data:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])

        yield TestRecord.deletesimple(txn, gamma=u"three")
        all = yield TestRecord.all(txn)
        self.assertEqual(set([record.beta for record in all]), set((123, 456, 234,)))

        yield TestRecord.deletesimple(txn, beta=123, gamma=u"one")
        all = yield TestRecord.all(txn)
        self.assertEqual(set([record.beta for record in all]), set((456, 234)))


    @inlineCallbacks
    def test_repr(self):
        """
        The C{repr} of a L{Record} presents all its values.
        """
        txn = self.pool.connection()
        yield txn.execSQL("insert into ALPHA values (:1, :2)", [789, u'nine'])
        rec = list((yield TestRecord.all(txn)))[0]
        self.assertIn(" beta=789", repr(rec))
        self.assertIn(" gamma=u'nine'", repr(rec))


    @inlineCallbacks
    def test_orderedQuery(self):
        """
        L{Record.query} takes an 'order' argument which will allow the objects
        returned to be ordered.
        """
        txn = self.pool.connection()
        for beta, gamma in [(123, u"one"), (234, u"two"), (345, u"three"),
                            (356, u"three"), (456, u"four")]:
            yield txn.execSQL("insert into ALPHA values (:1, :2)",
                              [beta, gamma])

        records = yield TestRecord.query(
            txn, TestRecord.gamma == u"three", TestRecord.beta
        )
        self.assertEqual([record.beta for record in records], [345, 356])

        records = yield TestRecord.query(
            txn, TestRecord.gamma == u"three", TestRecord.beta, ascending=False
        )
        self.assertEqual([record.beta for record in records], [356, 345])


    @inlineCallbacks
    def test_pop(self):
        """
        A L{Record} may be loaded and deleted atomically, with L{Record.pop}.
        """
        txn = self.pool.connection()
        for beta, gamma in [
            (123, u"one"),
            (234, u"two"),
            (345, u"three"),
            (356, u"three"),
            (456, u"four"),
        ]:
            yield txn.execSQL(
                "insert into ALPHA values (:1, :2)", [beta, gamma]
            )

        rec = yield TestRecord.pop(txn, 234)
        self.assertEqual(rec.gamma, u'two')
        self.assertEqual(
            (yield txn.execSQL(
                "select count(*) from ALPHA where BETA = :1", [234]
            )),
            [tuple([0])]
        )
        yield self.assertFailure(TestRecord.pop(txn, 234), NoSuchRecord)


    def test_columnNamingConvention(self):
        """
        The naming convention maps columns C{LIKE_THIS} to be attributes
        C{likeThis}.
        """
        self.assertEqual(
            Record.namingConvention(u"like_this"),
            "likeThis"
        )
        self.assertEqual(
            Record.namingConvention(u"LIKE_THIS"),
            "likeThis"
        )
        self.assertEqual(
            Record.namingConvention(u"LIKE_THIS_ID"),
            "likeThisID"
        )


    @inlineCallbacks
    def test_lock(self):
        """
        A L{Record} may be locked, with L{Record.lock}.
        """
        txn = self.pool.connection()
        for beta, gamma in [
            (123, u"one"),
            (234, u"two"),
            (345, u"three"),
            (356, u"three"),
            (456, u"four"),
        ]:
            yield txn.execSQL(
                "insert into ALPHA values (:1, :2)", [beta, gamma]
            )

        rec = yield TestRecord.load(txn, 234)
        yield rec.lock()
        self.assertEqual(rec.gamma, u'two')


    @inlineCallbacks
    def test_trylock(self):
        """
        A L{Record} may be locked, with L{Record.trylock}.
        """
        txn = self.pool.connection()
        for beta, gamma in [
            (123, u"one"),
            (234, u"two"),
            (345, u"three"),
            (356, u"three"),
            (456, u"four"),
        ]:
            yield txn.execSQL(
                "insert into ALPHA values (:1, :2)", [beta, gamma]
            )

        rec = yield TestRecord.load(txn, 234)
        result = yield rec.trylock()
        self.assertTrue(result)


    @inlineCallbacks
    def test_serialize(self):
        """
        A L{SerializableRecord} may be serialized.
        """
        txn = self.pool.connection()
        for beta, gamma in [
            (123, u"one"),
            (234, u"two"),
            (345, u"three"),
            (356, u"three"),
            (456, u"four"),
        ]:
            yield txn.execSQL(
                "insert into ALPHA values (:1, :2)", [beta, gamma]
            )

        rec = yield TestSerializeRecord.load(txn, 234)
        result = rec.serialize()
        self.assertEqual(result, {"beta": 234, "gamma": u"two"})


    @inlineCallbacks
    def test_deserialize(self):
        """
        A L{SerializableRecord} may be deserialized.
        """
        txn = self.pool.connection()

        rec = yield TestSerializeRecord.deserialize({"beta": 234, "gamma": u"two"})
        yield rec.insert(txn)
        yield txn.commit()

        txn = self.pool.connection()
        rec = yield TestSerializeRecord.query(txn, TestSerializeRecord.beta == 234)
        self.assertEqual(len(rec), 1)
        self.assertEqual(rec[0].gamma, u"two")
        yield txn.commit()

        # Check that attributes can be changed prior to insert, and not after
        txn = self.pool.connection()
        rec = yield TestSerializeRecord.deserialize({"beta": 456, "gamma": u"one"})
        rec.gamma = u"four"
        yield rec.insert(txn)
        def _raise():
            rec.gamma = u"five"
        self.assertRaises(ReadOnly, _raise)
        yield txn.commit()

        txn = self.pool.connection()
        rec = yield TestSerializeRecord.query(txn, TestSerializeRecord.beta == 456)
        self.assertEqual(len(rec), 1)
        self.assertEqual(rec[0].gamma, u"four")
        yield txn.commit()



##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Tests for L{twext.enterprise.dal.syntax}
"""

from twisted.internet.defer import succeed
from twisted.trial.unittest import TestCase, SkipTest

from twext.enterprise.dal import syntax
try:
    from twext.enterprise.dal.parseschema import addSQLToSchema
except ImportError as e:
    def addSQLToSchema(*args, **kwargs):
        raise SkipTest("addSQLToSchema is not available: {0}".format(e))
from twext.enterprise.dal.syntax import (
    Select, Insert, Update, Delete, Lock, SQLFragment,
    TableMismatch, Parameter, Max, Min, Len, NotEnoughValues,
    Savepoint, RollbackToSavepoint, ReleaseSavepoint, SavepointAction,
    Union, Intersect, Except, SetExpression, DALError,
    ResultAliasSyntax, Count, QueryGenerator, ALL_COLUMNS,
    DatabaseLock, DatabaseUnlock, Not, Coalesce)
from twext.enterprise.dal.syntax import FixedPlaceholder, NumericPlaceholder
from twext.enterprise.dal.syntax import Function
from twext.enterprise.dal.syntax import SchemaSyntax
from twext.enterprise.dal.test.test_parseschema import SchemaTestHelper
from twext.enterprise.ienterprise import (
    POSTGRES_DIALECT, ORACLE_DIALECT, SQLITE_DIALECT
)
from twext.enterprise.test.test_adbapi2 import ConnectionPoolHelper
from twext.enterprise.test.test_adbapi2 import NetworkedPoolHelper
from twext.enterprise.test.test_adbapi2 import resultOf, AssertResultHelper
from twext.enterprise.dal.syntax import Tuple
from twext.enterprise.dal.syntax import Constant



class _FakeTransaction(object):
    """
    An L{IAsyncTransaction} that provides the relevant metadata for SQL
    generation.
    """

    def __init__(self, paramstyle):
        self.paramstyle = "qmark"



class FakeCXOracleModule(object):
    NUMBER = "the NUMBER type"
    STRING = "a string type (for varchars)"
    NCLOB = "the NCLOB type. (for text)"
    TIMESTAMP = "for timestamps!"



class CatchSQL(object):
    """
    L{IAsyncTransaction} emulator that records the SQL executed on it.
    """
    counter = 0

    def __init__(self, dialect=SQLITE_DIALECT, paramstyle="numeric"):
        self.execed = []
        self.pendingResults = []
        self.dialect = SQLITE_DIALECT
        self.paramstyle = "numeric"


    def nextResult(self, result):
        """
        Make it so that the next result from L{execSQL} will be the argument.
        """
        self.pendingResults.append(result)


    def execSQL(self, sql, args, rozrc):
        """
        Implement L{IAsyncTransaction} by recording C{sql} and C{args} in
        C{self.execed}, and return a L{Deferred} firing either an integer or a
        value pre-supplied by L{CatchSQL.nextResult}.
        """
        self.execed.append([sql, args])
        self.counter += 1
        if self.pendingResults:
            result = self.pendingResults.pop(0)
        else:
            result = self.counter
        return succeed(result)



class NullTestingOracleTxn(object):
    """
    Fake transaction for testing oracle NULL behavior.
    """

    dialect = ORACLE_DIALECT
    paramstyle = "numeric"

    def execSQL(self, text, params, exc):
        return succeed([[None, None]])



EXAMPLE_SCHEMA = """
create sequence A_SEQ;
create table FOO (BAR integer, BAZ varchar(255));
create table BOZ (QUX integer, QUUX integer);
create table OTHER (BAR integer,
                    FOO_BAR integer not null);
create table TEXTUAL (MYTEXT varchar(255));
create table LEVELS (ACCESS integer,
                     USERNAME varchar(255));
create table NULLCHECK (ASTRING varchar(255) not null,
                        ANUMBER integer);
"""



class ExampleSchemaHelper(SchemaTestHelper):
    """
    setUp implementor.
    """

    def setUp(self):
        self.schema = SchemaSyntax(self.schemaFromString(EXAMPLE_SCHEMA))



class GenerationTests(ExampleSchemaHelper, TestCase, AssertResultHelper):
    """
    Tests for syntactic helpers to generate SQL queries.
    """

    def test_simplestSelect(self):
        """
        L{Select} generates a C{select} statement, by default, asking for all
        rows in a table.
        """
        self.assertEquals(
            Select(From=self.schema.FOO).toSQL(),
            SQLFragment("select * from FOO", [])
        )


    def test_tableSyntaxFromSchemaSyntaxCompare(self):
        """
        One L{TableSyntax} is equivalent to another wrapping the same table;
        one wrapping a different table is different.
        """
        self.assertEquals(self.schema.FOO, self.schema.FOO)
        self.assertNotEquals(self.schema.FOO, self.schema.BOZ)


    def test_simpleWhereClause(self):
        """
        L{Select} generates a C{select} statement with a C{where} clause
        containing an expression.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR == 1
            ).toSQL(),
            SQLFragment("select * from FOO where BAR = ?", [1])
        )


    def test_alternateMetadata(self):
        """
        L{Select} generates a C{select} statement with the specified
        placeholder syntax when explicitly given L{ConnectionMetadata} which
        specifies a placeholder.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR == 1
            ).toSQL(QueryGenerator(POSTGRES_DIALECT, FixedPlaceholder("$$"))),
            SQLFragment("select * from FOO where BAR = $$", [1])
        )


    def test_columnComparison(self):
        """
        L{Select} generates a C{select} statement which compares columns.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR == self.schema.FOO.BAZ
            ).toSQL(),
            SQLFragment("select * from FOO where BAR = BAZ", [])
        )


    def test_comparisonTestErrorPrevention(self):
        """
        The comparison object between SQL expressions raises an exception when
        compared for a truth value, so that code will not accidentally operate
        on SQL objects and get a truth value.

        (Note that this has a caveat, in test_columnsAsDictKeys and
        test_columnEqualityTruth.)
        """
        def sampleComparison():
            if self.schema.FOO.BAR > self.schema.FOO.BAZ:
                return "comparison should not succeed"
        self.assertRaises(DALError, sampleComparison)


    def test_compareWithNULL(self):
        """
        Comparing a column with None results in the generation of an C{is null}
        or C{is not null} SQL statement.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR == None
            ).toSQL(),
            SQLFragment("select * from FOO where BAR is null", [])
        )
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR != None
            ).toSQL(),
            SQLFragment("select * from FOO where BAR is not null", [])
        )


    def test_compareWithEmptyStringOracleSpecialCase(self):
        """
        Oracle considers the empty string to be a C{NULL} value, so comparisons
        with the empty string should be C{is NULL} comparisons.
        """
        # Sanity check: let's make sure that the non-oracle case looks normal.
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR == ""
            ).toSQL(),
            SQLFragment("select * from FOO where BAR = ?", [""])
        )
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR != ""
            ).toSQL(),
            SQLFragment("select * from FOO where BAR != ?", [""])
        )
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR == ""
            ).toSQL(QueryGenerator(ORACLE_DIALECT, NumericPlaceholder())),
            SQLFragment("select * from FOO where BAR is null", [])
        )
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR != ""
            ).toSQL(QueryGenerator(ORACLE_DIALECT, NumericPlaceholder())),
            SQLFragment("select * from FOO where BAR is not null", [])
        )


    def test_compoundWhere(self):
        """
        L{Select.And} and L{Select.Or} will return compound columns.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(self.schema.FOO.BAR < 2).Or(self.schema.FOO.BAR > 5)
            ).toSQL(),
            SQLFragment("select * from FOO where BAR < ? or BAR > ?", [2, 5])
        )


    def test_orderBy(self):
        """
        L{Select}'s L{OrderBy} parameter generates an C{order by} clause for a
        C{select} statement.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                OrderBy=self.schema.FOO.BAR
            ).toSQL(),
            SQLFragment("select * from FOO order by BAR")
        )


    def test_orderByOrder(self):
        """
        L{Select}'s L{Ascending} parameter specifies an ascending/descending
        order for query results with an OrderBy clause.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                OrderBy=self.schema.FOO.BAR,
                Ascending=False
            ).toSQL(),
            SQLFragment("select * from FOO order by BAR desc")
        )

        self.assertEquals(
            Select(
                From=self.schema.FOO,
                OrderBy=self.schema.FOO.BAR,
                Ascending=True
            ).toSQL(),
            SQLFragment("select * from FOO order by BAR asc")
        )

        self.assertEquals(
            Select(
                From=self.schema.FOO,
                OrderBy=[self.schema.FOO.BAR, self.schema.FOO.BAZ],
                Ascending=True
            ).toSQL(),
            SQLFragment("select * from FOO order by BAR, BAZ asc")
        )


    def test_orderByParens(self):
        """
        L{Select}'s L{OrderBy} paraneter, if specified as a L{Tuple}, generates
        an SQL expression I{without} parentheses, since the standard format
        does not allow an arbitrary sort expression but rather a list of
        columns.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                OrderBy=Tuple([self.schema.FOO.BAR, self.schema.FOO.BAZ])
            ).toSQL(),
            SQLFragment("select * from FOO order by BAR, BAZ")
        )


    def test_forUpdate(self):
        """
        L{Select}'s L{ForUpdate} parameter generates a C{for update} clause at
        the end of the query.
        """
        self.assertEquals(
            Select(From=self.schema.FOO, ForUpdate=True).toSQL(),
            SQLFragment("select * from FOO for update")
        )


    def test_groupBy(self):
        """
        L{Select}'s L{GroupBy} parameter generates a C{group by} clause for a
        C{select} statement.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                GroupBy=self.schema.FOO.BAR
            ).toSQL(),
            SQLFragment("select * from FOO group by BAR")
        )


    def test_groupByMulti(self):
        """
        L{Select}'s L{GroupBy} parameter can accept multiple columns in a list.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                GroupBy=[self.schema.FOO.BAR, self.schema.FOO.BAZ]
            ).toSQL(),
            SQLFragment("select * from FOO group by BAR, BAZ")
        )


    def test_joinClause(self):
        """
        A table's C{.join()} method returns a join statement in a C{SELECT}.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO.join(
                    self.schema.BOZ,
                    self.schema.FOO.BAR == self.schema.BOZ.QUX
                )
            ).toSQL(),
            SQLFragment("select * from FOO join BOZ on BAR = QUX", [])
        )


    def test_commaJoin(self):
        """
        A join with no clause specified will generate a cross join. This variant
        uses a "," between table names rather than "cross join".
        """
        self.assertEquals(
            Select(From=self.schema.FOO.join(self.schema.BOZ, type=",")).toSQL(),
            SQLFragment("select * from FOO, BOZ")
        )


    def test_crossJoin(self):
        """
        A join with no clause specified will generate a cross join.  (This is
        AN explicit synonym for an implicit join: i.e. C{select * from FOO,
        BAR}.)
        """
        self.assertEquals(
            Select(From=self.schema.FOO.join(self.schema.BOZ)).toSQL(),
            SQLFragment("select * from FOO cross join BOZ")
        )


    def test_joinJoin(self):
        """
        L{Join.join} will result in a multi-table join.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAR, self.schema.BOZ.QUX],
                From=self.schema.FOO
                .join(self.schema.BOZ)
                .join(self.schema.OTHER)
            ).toSQL(),
            SQLFragment(
                "select FOO.BAR, QUX from FOO cross join BOZ cross join OTHER"
            )
        )


    def test_multiJoin(self):
        """
        L{Join.join} has the same signature as L{TableSyntax.join} and supports
        the same C{on} and C{type} arguments.
        """

        self.assertEquals(
            Select(
                [self.schema.FOO.BAR],
                From=self.schema.FOO.join(
                    self.schema.BOZ
                ).join(
                    self.schema.OTHER,
                    self.schema.OTHER.BAR == self.schema.FOO.BAR,
                    "left outer"
                )
            ).toSQL(),
            SQLFragment(
                "select FOO.BAR from FOO cross join BOZ left outer join OTHER "
                "on OTHER.BAR = FOO.BAR"
            )
        )


    def test_tableAliasing(self):
        """
        Tables may be given aliases, in order to facilitate self-joins.
        """
        sfoo = self.schema.FOO
        sfoo2 = sfoo.alias()
        self.assertEqual(
            Select(From=self.schema.FOO.join(sfoo2)).toSQL(),
            SQLFragment("select * from FOO cross join FOO alias1")
        )


    def test_columnsOfAliasedTable(self):
        """
        The columns of aliased tables will always be prefixed with their alias
        in the generated SQL.
        """
        sfoo = self.schema.FOO
        sfoo2 = sfoo.alias()
        self.assertEquals(
            Select([sfoo2.BAR], From=sfoo2).toSQL(),
            SQLFragment("select alias1.BAR from FOO alias1")
        )


    def test_multipleTableAliases(self):
        """
        When multiple aliases are used for the same table, they will be unique
        within the query.
        """
        foo = self.schema.FOO
        fooPrime = foo.alias()
        fooPrimePrime = foo.alias()
        self.assertEquals(
            Select(
                [fooPrime.BAR, fooPrimePrime.BAR],
                From=fooPrime.join(fooPrimePrime)
            ).toSQL(),
            SQLFragment(
                "select alias1.BAR, alias2.BAR "
                "from FOO alias1 cross join FOO alias2"
            )
        )


    def test_columnSelection(self):
        """
        If a column is specified by the argument to L{Select}, those will be
        output by the SQL statement rather than the all-columns wildcard.
        """
        self.assertEquals(
            Select([self.schema.FOO.BAR], From=self.schema.FOO).toSQL(),
            SQLFragment("select BAR from FOO")
        )


    def test_tableIteration(self):
        """
        Iterating a L{TableSyntax} iterates its columns, in the order that they
        are defined.
        """
        self.assertEquals(
            list(self.schema.FOO),
            [self.schema.FOO.BAR, self.schema.FOO.BAZ]
        )


    def test_noColumn(self):
        """
        Accessing an attribute that is not a defined column on a L{TableSyntax}
        raises an L{AttributeError}.
        """
        self.assertRaises(AttributeError, lambda: self.schema.FOO.NOT_A_COLUMN)


    def test_columnAliases(self):
        """
        When attributes are set on a L{TableSyntax}, they will be remembered as
        column aliases, and their alias names may be retrieved via the
        L{TableSyntax.columnAliases} method.
        """
        self.assertEquals(self.schema.FOO.columnAliases(), {})
        self.schema.FOO.ALIAS = self.schema.FOO.BAR

        # you comparing ColumnSyntax object results in a ColumnComparison,
        # which you can't test for truth.
        fixedForEquality = dict([
            (k, v.model) for k, v in self.schema.FOO.columnAliases().items()
        ])

        self.assertEquals(
            fixedForEquality,
            {"ALIAS": self.schema.FOO.BAR.model}
        )
        self.assertIdentical(
            self.schema.FOO.ALIAS.model,
            self.schema.FOO.BAR.model
        )


    def test_multiColumnSelection(self):
        """
        If multiple columns are specified by the argument to L{Select}, those
        will be output by the SQL statement rather than the all-columns
        wildcard.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAZ, self.schema.FOO.BAR],
                From=self.schema.FOO
            ).toSQL(),
            SQLFragment("select BAZ, BAR from FOO")
        )


    def test_joinColumnSelection(self):
        """
        If multiple columns are specified by the argument to L{Select} that
        uses a L{TableSyntax.join}, those will be output by the SQL statement.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAZ, self.schema.BOZ.QUX],
                From=self.schema.FOO.join(
                    self.schema.BOZ,
                    self.schema.FOO.BAR == self.schema.BOZ.QUX
                )
            ).toSQL(),
            SQLFragment("select BAZ, QUX from FOO join BOZ on BAR = QUX")
        )


    def test_tableMismatch(self):
        """
        When a column in the C{columns} argument does not match the table from
        the C{From} argument, L{Select} raises a L{TableMismatch}.
        """
        self.assertRaises(
            TableMismatch,
            Select, [self.schema.BOZ.QUX], From=self.schema.FOO
        )


    def test_qualifyNames(self):
        """
        When two columns in the C{from} clause requested from different tables
        have the same name, the emitted SQL should explicitly disambiguate
        them.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAR, self.schema.OTHER.BAR],
                From=self.schema.FOO.join(
                    self.schema.OTHER,
                    self.schema.OTHER.FOO_BAR == self.schema.FOO.BAR
                )
            ).toSQL(),
            SQLFragment(
                "select FOO.BAR, OTHER.BAR from FOO "
                "join OTHER on FOO_BAR = FOO.BAR"
            )
        )


    def test_bindParameters(self):
        """
        L{SQLFragment.bind} returns a copy of that L{SQLFragment} with the
        L{Parameter} objects in its parameter list replaced with the keyword
        arguments to C{bind}.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(self.schema.FOO.BAR > Parameter("testing"))
                .And(self.schema.FOO.BAZ < 7)
            ).toSQL().bind(testing=173),
            SQLFragment(
                "select * from FOO where BAR > ? and BAZ < ?", [173, 7]
            )
        )


    def test_rightHandSideExpression(self):
        """
        Arbitrary expressions may be used as the right-hand side of a
        comparison operation.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR > (self.schema.FOO.BAZ + 3)
            ).toSQL(),
            SQLFragment("select * from FOO where BAR > (BAZ + ?)", [3])
        )


    def test_setSelects(self):
        """
        L{SetExpression} produces set operation on selects.
        """
        # Simple UNION
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(self.schema.FOO.BAR == 1),
                SetExpression=Union(
                    Select(
                        From=self.schema.FOO,
                        Where=(self.schema.FOO.BAR == 2),
                    ),
                ),
            ).toSQL(QueryGenerator(POSTGRES_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                "(select * from FOO where BAR = ?) "
                "UNION (select * from FOO where BAR = ?)", [1, 2]
            )
        )

        # Simple INTERSECT ALL
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(self.schema.FOO.BAR == 1),
                SetExpression=Intersect(
                    Select(
                        From=self.schema.FOO,
                        Where=(self.schema.FOO.BAR == 2),
                    ),
                    optype=SetExpression.OPTYPE_ALL
                ),
            ).toSQL(QueryGenerator(POSTGRES_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                "(select * from FOO where BAR = ?) "
                "INTERSECT ALL (select * from FOO where BAR = ?)", [1, 2]
            )
        )

        # Multiple EXCEPTs, not nested, Postgres dialect
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                SetExpression=Except(
                    (
                        Select(
                            From=self.schema.FOO,
                            Where=(self.schema.FOO.BAR == 2),
                        ),
                        Select(
                            From=self.schema.FOO,
                            Where=(self.schema.FOO.BAR == 3),
                        ),
                    ),
                    optype=SetExpression.OPTYPE_DISTINCT,
                ),
            ).toSQL(QueryGenerator(POSTGRES_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                "(select * from FOO) "
                "EXCEPT DISTINCT (select * from FOO where BAR = ?) "
                "EXCEPT DISTINCT (select * from FOO where BAR = ?)", [2, 3]
            )
        )

        # Nested EXCEPTs, Oracle dialect
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                SetExpression=Except(
                    Select(
                        From=self.schema.FOO,
                        Where=(self.schema.FOO.BAR == 2),
                        SetExpression=Except(
                            Select(
                                From=self.schema.FOO,
                                Where=(self.schema.FOO.BAR == 3),
                            ),
                        ),
                    ),
                ),
            ).toSQL(QueryGenerator(ORACLE_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                "(select * from FOO) MINUS ((select * from FOO where BAR = ?) "
                "MINUS (select * from FOO where BAR = ?))", [2, 3]
            )
        )

        # UNION with order by
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(self.schema.FOO.BAR == 1),
                SetExpression=Union(
                    Select(
                        From=self.schema.FOO,
                        Where=(self.schema.FOO.BAR == 2),
                    ),
                ),
                OrderBy=self.schema.FOO.BAR,
            ).toSQL(QueryGenerator(POSTGRES_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                "(select * from FOO where BAR = ?) "
                "UNION (select * from FOO where BAR = ?) order by BAR", [1, 2]
            )
        )


    def test_simpleSubSelects(self):
        """
        L{Max}C{(column)} produces an object in the C{columns} clause that
        renders the C{max} aggregate in SQL.
        """
        self.assertEquals(
            Select(
                [Max(self.schema.BOZ.QUX)],
                From=(Select([self.schema.BOZ.QUX], From=self.schema.BOZ))
            ).toSQL(),
            SQLFragment("select max(QUX) from (select QUX from BOZ) genid_1")
        )

        self.assertEquals(
            Select(
                [Count(self.schema.BOZ.QUX)],
                From=(Select([self.schema.BOZ.QUX], From=self.schema.BOZ))
            ).toSQL(),
            SQLFragment("select count(QUX) from (select QUX from BOZ) genid_1")
        )

        self.assertEquals(
            Select(
                [Max(self.schema.BOZ.QUX)],
                From=(Select(
                    [self.schema.BOZ.QUX],
                    From=self.schema.BOZ,
                    As="alias_BAR"
                )),
            ).toSQL(),
            SQLFragment("select max(QUX) from (select QUX from BOZ) alias_BAR")
        )


    def test_setSubSelects(self):
        """
        L{SetExpression} in a C{From} sub-select.
        """
        # Simple UNION
        self.assertEquals(
            Select(
                [Max(self.schema.FOO.BAR)],
                From=Select(
                    [self.schema.FOO.BAR],
                    From=self.schema.FOO,
                    Where=(self.schema.FOO.BAR == 1),
                    SetExpression=Union(
                        Select(
                            [self.schema.FOO.BAR],
                            From=self.schema.FOO,
                            Where=(self.schema.FOO.BAR == 2),
                        ),
                    ),
                )
            ).toSQL(),
            SQLFragment(
                "select max(BAR) from ((select BAR from FOO where BAR = ?) "
                "UNION (select BAR from FOO where BAR = ?)) genid_1", [1, 2]
            )
        )


    def test_selectColumnAliases(self):
        """
        L{Select} works with aliased columns.
        """
        self.assertEquals(
            Select(
                [ResultAliasSyntax(self.schema.BOZ.QUX, "BOZ_QUX")],
                From=self.schema.BOZ
            ).toSQL(),
            SQLFragment("select QUX BOZ_QUX from BOZ")
        )

        self.assertEquals(
            Select(
                [ResultAliasSyntax(Max(self.schema.BOZ.QUX))],
                From=self.schema.BOZ
            ).toSQL(),
            SQLFragment("select max(QUX) genid_1 from BOZ")
        )

        alias = ResultAliasSyntax(Max(self.schema.BOZ.QUX))
        self.assertEquals(
            Select(
                [alias.columnReference()],
                From=Select([alias], From=self.schema.BOZ)
            ).toSQL(),
            SQLFragment(
                "select genid_1 from "
                "(select max(QUX) genid_1 from BOZ) genid_2"
            )
        )

        alias = ResultAliasSyntax(Len(self.schema.BOZ.QUX))
        self.assertEquals(
            Select(
                [alias.columnReference()],
                From=Select([alias], From=self.schema.BOZ)
            ).toSQL(),
            SQLFragment(
                "select genid_1 from "
                "(select character_length(QUX) genid_1 from BOZ) genid_2"
            )
        )


    def test_inSubSelect(self):
        """
        L{ColumnSyntax.In} returns a sub-expression using the SQL C{in} syntax
        with a sub-select.
        """
        wherein = self.schema.FOO.BAR.In(
            Select([self.schema.BOZ.QUX], From=self.schema.BOZ)
        )
        self.assertEquals(
            Select(From=self.schema.FOO, Where=wherein).toSQL(),
            SQLFragment(
                "select * from FOO where BAR in (select QUX from BOZ)"
            )
        )


    def test_inParameter(self):
        """
        L{ColumnSyntax.In} returns a sub-expression using the SQL C{in} syntax
        with parameter list.
        """
        # One item with IN only
        items = set(("A",))
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR.In(
                    Parameter("names", len(items))
                )
            ).toSQL().bind(names=items),
            SQLFragment("select * from FOO where BAR in (?)", ["A"])
        )

        # Two items with IN only
        items = set(("A", "B"))
        test = Select(
            From=self.schema.FOO,
            Where=self.schema.FOO.BAR.In(
                Parameter("names", len(items))
            )
        ).toSQL().bind(names=items)
        test.parameters.sort()
        self.assertEquals(
            test,
            SQLFragment(
                "select * from FOO where BAR in (?, ?)", ["A", "B"]
            )
        )

        # Two items with preceding AND
        test = Select(
            From=self.schema.FOO,
            Where=(
                (
                    self.schema.FOO.BAZ == Parameter("P1")
                ).And(
                    self.schema.FOO.BAR.In(Parameter("names", len(items)))
                )
            )
        ).toSQL().bind(P1="P1", names=items)
        test.parameters = [test.parameters[0], ] + sorted(test.parameters[1:])
        self.assertEquals(
            test,
            SQLFragment(
                "select * from FOO where BAZ = ? and BAR in (?, ?)",
                ["P1", "A", "B"]
            ),
        )

        # Two items with following AND
        test = Select(
            From=self.schema.FOO,
            Where=(
                (
                    self.schema.FOO.BAR.In(Parameter("names", len(items)))
                ).And(
                    self.schema.FOO.BAZ == Parameter("P2")
                )
            )
        ).toSQL().bind(P2="P2", names=items)
        test.parameters = sorted(test.parameters[:-1]) + [test.parameters[-1], ]
        self.assertEquals(
            test,
            SQLFragment(
                "select * from FOO where BAR in (?, ?) and BAZ = ?",
                ["A", "B", "P2"]
            ),
        )

        # Two items with preceding OR and following AND
        test = Select(
            From=self.schema.FOO,
            Where=((
                self.schema.FOO.BAZ == Parameter("P1")
            ).Or(
                self.schema.FOO.BAR.In(Parameter("names", len(items))).And(
                    self.schema.FOO.BAZ == Parameter("P2")
                )
            ))
        ).toSQL().bind(P1="P1", P2="P2", names=items)
        test.parameters = [test.parameters[0], ] + sorted(test.parameters[1:-1]) + [test.parameters[-1], ]
        self.assertEquals(
            test,
            SQLFragment(
                "select * from FOO where BAZ = ? or BAR in (?, ?) and BAZ = ?",
                ["P1", "A", "B", "P2"]
            ),
        )

        # Check various error situations

        # No count not allowed
        self.assertRaises(DALError, self.schema.FOO.BAR.In, Parameter("names"))

        # count=0 not allowed
        self.assertRaises(DALError, Parameter, "names", 0)

        # Mismatched count and len(items)
        self.assertRaises(
            DALError,
            Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR.In(Parameter("names", len(items)))
            ).toSQL().bind,
            names=["a", "b", "c"]
        )


    def test_inIterable(self):
        """
        L{ColumnSyntax.In} returns a sub-expression using the SQL C{in} syntax
        with parameter list.
        """
        # One item with IN only
        items = set(("A",))
        for items in (set(items), list(items), tuple(items),):
            self.assertEquals(
                Select(
                    From=self.schema.FOO,
                    Where=self.schema.FOO.BAR.In(items),
                ).toSQL().bind(),
                SQLFragment("select * from FOO where BAR in (?)", ["A"])
            )

        # Two items with IN only
        items = ("A", "B")
        for items in (set(items), list(items), tuple(items),):
            test = Select(
                From=self.schema.FOO,
                Where=self.schema.FOO.BAR.In(items),
            ).toSQL().bind(names=items)
            test.parameters.sort()
            self.assertEquals(
                test,
                SQLFragment(
                    "select * from FOO where BAR in (?, ?)", ["A", "B"]
                )
            )

            # Two items with preceding AND
            test = Select(
                From=self.schema.FOO,
                Where=(
                    (
                        self.schema.FOO.BAZ == Parameter("P1")
                    ).And(
                        self.schema.FOO.BAR.In(items)
                    )
                )
            ).toSQL().bind(P1="P1")
            test.parameters = [test.parameters[0], ] + sorted(test.parameters[1:])
            self.assertEquals(
                test,
                SQLFragment(
                    "select * from FOO where BAZ = ? and BAR in (?, ?)",
                    ["P1", "A", "B"]
                ),
            )

            # Two items with following AND
            test = Select(
                From=self.schema.FOO,
                Where=(
                    (
                        self.schema.FOO.BAR.In(items)
                    ).And(
                        self.schema.FOO.BAZ == Parameter("P2")
                    )
                )
            ).toSQL().bind(P2="P2")
            test.parameters = sorted(test.parameters[:-1]) + [test.parameters[-1], ]
            self.assertEquals(
                test,
                SQLFragment(
                    "select * from FOO where BAR in (?, ?) and BAZ = ?",
                    ["A", "B", "P2"]
                ),
            )

            # Two items with preceding OR and following AND
            test = Select(
                From=self.schema.FOO,
                Where=((
                    self.schema.FOO.BAZ == Parameter("P1")
                ).Or(
                    self.schema.FOO.BAR.In(items).And(
                        self.schema.FOO.BAZ == Parameter("P2")
                    )
                ))
            ).toSQL().bind(P1="P1", P2="P2")
            test.parameters = [test.parameters[0], ] + sorted(test.parameters[1:-1]) + [test.parameters[-1], ]
            self.assertEquals(
                test,
                SQLFragment(
                    "select * from FOO where BAZ = ? or BAR in (?, ?) and BAZ = ?",
                    ["P1", "A", "B", "P2"]
                ),
            )


    def test_max(self):
        """
        L{Max}C{(column)} produces an object in the C{columns} clause that
        renders the C{max} aggregate in SQL.
        """
        self.assertEquals(
            Select([Max(self.schema.BOZ.QUX)], From=self.schema.BOZ).toSQL(),
            SQLFragment("select max(QUX) from BOZ")
        )


    def test_min(self):
        """
        L{Min}C{(column)} produces an object in the C{columns} clause that
        renders the C{min} aggregate in SQL.
        """
        self.assertEquals(
            Select([Min(self.schema.BOZ.QUX)], From=self.schema.BOZ).toSQL(),
            SQLFragment("select min(QUX) from BOZ")
        )


    def test_coalesce(self):
        """
        L{Coalesce}C{(column)} produces an object in the C{columns} clause that
        renders the C{coalesce} conditional in SQL.
        """
        self.assertEquals(
            Select(
                [self.schema.BOZ.QUX],
                From=self.schema.BOZ,
                Where=Coalesce(self.schema.BOZ.QUX, self.schema.BOZ.QUUX) == 1
            ).toSQL(),
            SQLFragment("select QUX from BOZ where coalesce(QUX, QUUX) = ?", [1])
        )


    def test_countAllCoumns(self):
        """
        L{Count}C{(ALL_COLUMNS)} produces an object in the C{columns} clause
        that renders the C{count} in SQL.
        """
        self.assertEquals(
            Select([Count(ALL_COLUMNS)], From=self.schema.BOZ).toSQL(),
            SQLFragment("select count(*) from BOZ")
        )


    def test_aggregateComparison(self):
        """
        L{Max}C{(column) > constant} produces an object in the C{columns}
        clause that renders a comparison to the C{max} aggregate in SQL.
        """
        self.assertEquals(
            Select(
                [Max(self.schema.BOZ.QUX) + 12],
                From=self.schema.BOZ
            ).toSQL(),
            SQLFragment("select max(QUX) + ? from BOZ", [12])
        )


    def test_multiColumnExpression(self):
        """
        Multiple columns may be provided in an expression in the C{columns}
        portion of a C{Select()} statement.  All arithmetic operators are
        supported.
        """
        self.assertEquals(
            Select(
                [((self.schema.FOO.BAR + self.schema.FOO.BAZ) / 3) * 7],
                From=self.schema.FOO
            ).toSQL(),
            SQLFragment("select ((BAR + BAZ) / ?) * ? from FOO", [3, 7])
        )


    def test_len(self):
        """
        Test for the L{Len} function for determining character length of a
        column.  (Note that this should be updated to use different techniques
        as necessary in different databases.)
        """
        self.assertEquals(
            Select(
                [Len(self.schema.TEXTUAL.MYTEXT)],
                From=self.schema.TEXTUAL
            ).toSQL(),
            SQLFragment("select character_length(MYTEXT) from TEXTUAL")
        )


    def test_startswith(self):
        """
        Test for the string starts with comparison.
        (Note that this should be updated to use different techniques
        as necessary in different databases.)
        """
        self.assertEquals(
            Select(
                [self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=self.schema.TEXTUAL.MYTEXT.StartsWith("test"),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where MYTEXT like (? || ?)",
                ["test", "%"]
            )
        )


    def test_endswith(self):
        """
        Test for the string starts with comparison.
        (Note that this should be updated to use different techniques
        as necessary in different databases.)
        """
        self.assertEquals(
            Select(
                [self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=self.schema.TEXTUAL.MYTEXT.EndsWith("test"),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where MYTEXT like (? || ?)",
                ["%", "test"]
            )
        )


    def test_contains(self):
        """
        Test for the string starts with comparison.
        (Note that this should be updated to use different techniques
        as necessary in different databases.)
        """
        self.assertEquals(
            Select(
                [self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=self.schema.TEXTUAL.MYTEXT.Contains("test"),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where MYTEXT like (? || (? || ?))",
                ["%", "test", "%"]
            )
        )


    def test_not(self):
        """
        Test for the string starts with comparison.
        (Note that this should be updated to use different techniques
        as necessary in different databases.)
        """
        self.assertEquals(
            Select([
                self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=Not(self.schema.TEXTUAL.MYTEXT.StartsWith("test")),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where not MYTEXT like (? || ?)",
                ["test", "%"]
            )
        )

        self.assertEquals(
            Select([
                self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=Not(self.schema.TEXTUAL.MYTEXT == "test"),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where not MYTEXT = ?",
                ["test"]
            )
        )

        self.assertEquals(
            Select([
                self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=Not((self.schema.TEXTUAL.MYTEXT == "test1").And(self.schema.TEXTUAL.MYTEXT != "test2")),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where not (MYTEXT = ? and MYTEXT != ?)",
                ["test1", "test2"]
            )
        )

        self.assertEquals(
            Select([
                self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=Not((self.schema.TEXTUAL.MYTEXT == "test1")).And(self.schema.TEXTUAL.MYTEXT != "test2"),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where not MYTEXT = ? and MYTEXT != ?",
                ["test1", "test2"]
            )
        )

        self.assertEquals(
            Select([
                self.schema.TEXTUAL.MYTEXT],
                From=self.schema.TEXTUAL,
                Where=Not(self.schema.TEXTUAL.MYTEXT.StartsWith("foo").And(self.schema.TEXTUAL.MYTEXT.NotEndsWith("bar"))),
            ).toSQL(),
            SQLFragment(
                "select MYTEXT from TEXTUAL where not (MYTEXT like (? || ?) and MYTEXT not like (? || ?))",
                ["foo", "%", "%", "bar"]
            )
        )


    def test_insert(self):
        """
        L{Insert.toSQL} generates an C{insert} statement with all the relevant
        columns.
        """
        self.assertEquals(
            Insert(
                {self.schema.FOO.BAR: 23, self.schema.FOO.BAZ: 9}
            ).toSQL(),
            SQLFragment("insert into FOO (BAR, BAZ) values (?, ?)", [23, 9])
        )


    def test_insertNotEnough(self):
        """
        L{Insert}'s constructor will raise L{NotEnoughValues} if columns have
        not been specified.
        """
        notEnough = self.assertRaises(
            NotEnoughValues, Insert, {self.schema.OTHER.BAR: 9}
        )
        self.assertEquals(str(notEnough), "Columns [FOO_BAR] required.")


    def test_insertReturning(self):
        """
        L{Insert}'s C{Return} argument will insert an SQL C{returning} clause.
        """
        self.assertEquals(
            Insert(
                {self.schema.FOO.BAR: 23, self.schema.FOO.BAZ: 9},
                Return=self.schema.FOO.BAR
            ).toSQL(),
            SQLFragment(
                "insert into FOO (BAR, BAZ) values (?, ?) returning BAR",
                [23, 9]
            )
        )


    def test_insertMultiReturn(self):
        """
        L{Insert}'s C{Return} argument can also be a C{tuple}, which will
        insert an SQL C{returning} clause with multiple columns.
        """
        self.assertEquals(
            Insert(
                {self.schema.FOO.BAR: 23, self.schema.FOO.BAZ: 9},
                Return=(self.schema.FOO.BAR, self.schema.FOO.BAZ)
            ).toSQL(),
            SQLFragment(
                "insert into FOO (BAR, BAZ) values (?, ?) returning BAR, BAZ",
                [23, 9]
            )
        )


    def test_insertMultiReturnOracle(self):
        """
        In Oracle's SQL dialect, the C{returning} clause requires an C{into}
        clause indicating where to put the results, as they can't be simply
        relayed to the cursor.  Further, additional bound variables are
        required to capture the output parameters.
        """
        self.assertEquals(
            Insert(
                {self.schema.FOO.BAR: 40, self.schema.FOO.BAZ: 50},
                Return=(self.schema.FOO.BAR, self.schema.FOO.BAZ)
            ).toSQL(QueryGenerator(ORACLE_DIALECT, NumericPlaceholder())),
            SQLFragment(
                "insert into FOO (BAR, BAZ) values (:1, :2) "
                "returning BAR, BAZ into :3, :4",
                [40, 50, Parameter("oracle_out_0"), Parameter("oracle_out_1")]
            )
        )


    def test_insertMultiReturnSQLite(self):
        """
        In SQLite's SQL dialect, there is no C{returning} clause, but given
        that SQLite serializes all SQL transactions, you can rely upon
        C{select} after a write operation to reliably give you exactly what was
        just modified.  Therefore, although C{toSQL} won't include any
        indication of the return value, the C{on} method will execute a
        C{select} statement following the insert to retrieve the value.
        """
        insertStatement = Insert(
            {self.schema.FOO.BAR: 39, self.schema.FOO.BAZ: 82},
            Return=(self.schema.FOO.BAR, self.schema.FOO.BAZ)
        )
        qg = lambda: QueryGenerator(SQLITE_DIALECT, NumericPlaceholder())
        self.assertEquals(
            insertStatement.toSQL(qg()),
            SQLFragment("insert into FOO (BAR, BAZ) values (:1, :2)", [39, 82])
        )
        result = []
        csql = CatchSQL()
        insertStatement.on(csql).addCallback(result.append)
        self.assertEqual(result, [2])
        self.assertEqual(
            csql.execed,
            [
                [
                    "insert into FOO (BAR, BAZ) values (:1, :2)",
                    [39, 82]
                ],
                [
                    "select BAR, BAZ from FOO "
                    "where rowid = last_insert_rowid()",
                    []
                ],
            ]
        )


    def test_insertNoReturnSQLite(self):
        """
        Insert a row I{without} a C{Return=} parameter should also work as
        normal in sqlite.
        """
        statement = Insert(
            {self.schema.FOO.BAR: 12, self.schema.FOO.BAZ: 48}
        )
        csql = CatchSQL()
        statement.on(csql)
        self.assertEqual(
            csql.execed,
            [["insert into FOO (BAR, BAZ) values (:1, :2)", [12, 48]]]
        )


    def test_updateReturningSQLite(self):
        """
        Since SQLite does not support the SQL C{returning} syntax extension, in
        order to preserve the rows that will be modified during an UPDATE
        statement, we must first find the rows that will be affected, then
        update them, then return the rows that were affected.  Since we might
        be changing even part of the primary key, we use the internal C{rowid}
        column to uniquely and reliably identify rows in the sqlite database
        that have been modified.
        """
        csql = CatchSQL()
        stmt = Update(
            {self.schema.FOO.BAR: 4321},
            Where=self.schema.FOO.BAZ == 1234,
            Return=self.schema.FOO.BAR
        )
        csql.nextResult([["sample row id"]])
        result = resultOf(stmt.on(csql))

        # Three statements were executed; make sure that the result returned
        # was the result of executing the 3rd (and final) one.
        self.assertResultList(result, 3)

        # Check that they were the right statements.
        self.assertEqual(len(csql.execed), 3)
        self.assertEqual(
            csql.execed[0],
            ["select rowid from FOO where BAZ = :1", [1234]]
        )
        self.assertEqual(
            csql.execed[1],
            ["update FOO set BAR = :1 where BAZ = :2", [4321, 1234]]
        )
        self.assertEqual(
            csql.execed[2],
            ["select BAR from FOO where rowid = :1", ["sample row id"]]
        )


    def test_updateReturningMultipleValuesSQLite(self):
        """
        When SQLite updates multiple values, it must embed the row ID of each
        subsequent value into its second C{where} clause, as there is no way to
        pass a list of values to a single statement..
        """
        csql = CatchSQL()
        stmt = Update(
            {self.schema.FOO.BAR: 4321},
            Where=self.schema.FOO.BAZ == 1234,
            Return=self.schema.FOO.BAR
        )
        csql.nextResult([["one row id"], ["and another"], ["and one more"]])
        result = resultOf(stmt.on(csql))

        # Three statements were executed; make sure that the result returned
        # was the result of executing the 3rd (and final) one.
        self.assertResultList(result, 3)

        # Check that they were the right statements.
        self.assertEqual(len(csql.execed), 3)
        self.assertEqual(
            csql.execed[0],
            ["select rowid from FOO where BAZ = :1", [1234]]
        )
        self.assertEqual(
            csql.execed[1],
            ["update FOO set BAR = :1 where BAZ = :2", [4321, 1234]]
        )
        self.assertEqual(
            csql.execed[2],
            [
                "select BAR from FOO "
                "where rowid = :1 or rowid = :2 or rowid = :3",
                ["one row id", "and another", "and one more"]
            ]
        )


    def test_deleteReturningSQLite(self):
        """
        When SQLite deletes a value, ...
        """
        csql = CatchSQL()
        stmt = Delete(
            From=self.schema.FOO,
            Where=self.schema.FOO.BAZ == 1234,
            Return=self.schema.FOO.BAR
        )
        result = resultOf(stmt.on(csql))
        self.assertResultList(result, 1)
        self.assertEqual(len(csql.execed), 2)
        self.assertEqual(
            csql.execed[0],
            ["select BAR from FOO where BAZ = :1", [1234]]
        )
        self.assertEqual(
            csql.execed[1],
            ["delete from FOO where BAZ = :1", [1234]]
        )


    def test_insertMismatch(self):
        """
        L{Insert} raises L{TableMismatch} if the columns specified aren't all
        from the same table.
        """
        self.assertRaises(
            TableMismatch,
            Insert, {
                self.schema.FOO.BAR: 23,
                self.schema.FOO.BAZ: 9,
                self.schema.TEXTUAL.MYTEXT: "hello"
            }
        )


    def test_quotingOnKeywordConflict(self):
        """
        "access" is a keyword, so although our schema parser will leniently
        accept it, it must be quoted in any outgoing SQL.  (This is only done
        in the Oracle dialect, because it isn't necessary in postgres, and
        idiosyncratic case-folding rules make it challenging to do it in both.)
        """
        self.assertEquals(
            Insert(
                {
                    self.schema.LEVELS.ACCESS: 1,
                    self.schema.LEVELS.USERNAME: "hi"
                }
            ).toSQL(QueryGenerator(ORACLE_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                """insert into LEVELS ("ACCESS", USERNAME) values (?, ?)""",
                [1, "hi"]
            )
        )
        self.assertEquals(
            Insert(
                {
                    self.schema.LEVELS.ACCESS: 1,
                    self.schema.LEVELS.USERNAME: "hi"
                }
            ).toSQL(QueryGenerator(POSTGRES_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                "insert into LEVELS (ACCESS, USERNAME) values (?, ?)",
                [1, "hi"]
            )
        )


    def test_updateReturning(self):
        """
        L{update}'s C{Return} argument will update an SQL C{returning} clause.
        """
        self.assertEquals(
            Update(
                {self.schema.FOO.BAR: 23},
                self.schema.FOO.BAZ == 43,
                Return=self.schema.FOO.BAR
            ).toSQL(),
            SQLFragment(
                "update FOO set BAR = ? where BAZ = ? returning BAR",
                [23, 43]
            )
        )


    def test_updateMismatch(self):
        """
        L{Update} raises L{TableMismatch} if the columns specified aren't all
        from the same table.
        """
        self.assertRaises(
            TableMismatch,
            Update, {
                self.schema.FOO.BAR: 23,
                self.schema.FOO.BAZ: 9,
                self.schema.TEXTUAL.MYTEXT: "hello"
            },
            Where=self.schema.FOO.BAZ == 9
        )


    def test_updateFunction(self):
        """
        L{Update} values may be L{FunctionInvocation}s, to update to computed
        values in the database.
        """
        sqlfunc = Function("hello")
        self.assertEquals(
            Update(
                {self.schema.FOO.BAR: 23, self.schema.FOO.BAZ: sqlfunc()},
                Where=self.schema.FOO.BAZ == 9
            ).toSQL(),
            SQLFragment(
                "update FOO set BAR = ?, BAZ = hello() where BAZ = ?", [23, 9]
            )
        )


    def test_insertFunction(self):
        """
        L{Update} values may be L{FunctionInvocation}s, to update to computed
        values in the database.
        """
        sqlfunc = Function("hello")
        self.assertEquals(
            Insert(
                {self.schema.FOO.BAR: 23, self.schema.FOO.BAZ: sqlfunc()},
            ).toSQL(),
            SQLFragment("insert into FOO (BAR, BAZ) values (?, hello())", [23])
        )


    def test_deleteReturning(self):
        """
        L{Delete}'s C{Return} argument will delete an SQL C{returning} clause.
        """
        self.assertEquals(
            Delete(
                self.schema.FOO,
                Where=self.schema.FOO.BAR == 7,
                Return=self.schema.FOO.BAZ
            ).toSQL(),
            SQLFragment("delete from FOO where BAR = ? returning BAZ", [7])
        )


    def test_update(self):
        """
        L{Update.toSQL} generates an C{update} statement.
        """
        self.assertEquals(
            Update(
                {self.schema.FOO.BAR: 4321},
                self.schema.FOO.BAZ == 1234
            ).toSQL(),
            SQLFragment("update FOO set BAR = ? where BAZ = ?", [4321, 1234])
        )


    def test_delete(self):
        """
        L{Delete} generates an SQL C{delete} statement.
        """
        self.assertEquals(
            Delete(self.schema.FOO, Where=self.schema.FOO.BAR == 12).toSQL(),
            SQLFragment("delete from FOO where BAR = ?", [12])
        )

        self.assertEquals(
            Delete(self.schema.FOO, Where=None).toSQL(),
            SQLFragment("delete from FOO")
        )


    def test_lock(self):
        """
        L{Lock.exclusive} generates a C{lock table} statement, locking the
        table in the specified mode.
        """
        self.assertEquals(
            Lock.exclusive(self.schema.FOO).toSQL(),
            SQLFragment("lock table FOO in exclusive mode")
        )


    def test_databaseLock(self):
        """
        L{DatabaseLock} generates a C{pg_advisory_lock} statement
        """
        self.assertEquals(
            DatabaseLock().toSQL(),
            SQLFragment("select pg_advisory_lock(1)")
        )


    def test_databaseUnlock(self):
        """
        L{DatabaseUnlock} generates a C{pg_advisory_unlock} statement
        """
        self.assertEquals(
            DatabaseUnlock().toSQL(),
            SQLFragment("select pg_advisory_unlock(1)")
        )


    def test_savepoint(self):
        """
        L{Savepoint} generates a C{savepoint} statement.
        """
        self.assertEquals(
            Savepoint("test").toSQL(),
            SQLFragment("savepoint test")
        )


    def test_rollbacktosavepoint(self):
        """
        L{RollbackToSavepoint} generates a C{rollback to savepoint} statement.
        """
        self.assertEquals(
            RollbackToSavepoint("test").toSQL(),
            SQLFragment("rollback to savepoint test")
        )


    def test_releasesavepoint(self):
        """
        L{ReleaseSavepoint} generates a C{release savepoint} statement.
        """
        self.assertEquals(
            ReleaseSavepoint("test").toSQL(),
            SQLFragment("release savepoint test")
        )


    def test_savepointaction(self):
        """
        L{SavepointAction} generates a C{savepoint} statement.
        """
        self.assertEquals(SavepointAction("test")._name, "test")


    def test_limit(self):
        """
        A L{Select} object with a C{Limit} keyword parameter will generate
        a SQL statement with a C{limit} clause.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAR],
                From=self.schema.FOO,
                Limit=123
            ).toSQL(),
            SQLFragment("select BAR from FOO limit ?", [123])
        )


    def test_limitOracle(self):
        """
        A L{Select} object with a C{Limit} keyword parameter will generate a
        SQL statement using a ROWNUM subquery for Oracle.

        See U{this "ask tom" article from 2006 for more
        information
        <http://www.oracle.com/technetwork/issue-archive/2006/06-sep/o56asktom-086197.html>}.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAR],
                From=self.schema.FOO,
                Limit=123
            ).toSQL(QueryGenerator(ORACLE_DIALECT, FixedPlaceholder("?"))),
            SQLFragment(
                "select * from (select BAR from FOO) "
                "where ROWNUM <= ?", [123]
            )
        )


    def test_having(self):
        """
        A L{Select} object with a C{Having} keyword parameter will generate
        a SQL statement with a C{having} expression.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAR],
                From=self.schema.FOO,
                Having=Max(self.schema.FOO.BAZ) < 7
            ).toSQL(),
            SQLFragment("select BAR from FOO having max(BAZ) < ?", [7])
        )


    def test_distinct(self):
        """
        A L{Select} object with a C{Disinct} keyword parameter with a value of
        C{True} will generate a SQL statement with a C{distinct} keyword
        preceding its list of columns.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAR],
                From=self.schema.FOO,
                Distinct=True
            ).toSQL(),
            SQLFragment("select distinct BAR from FOO")
        )


    def test_nextSequenceValue(self):
        """
        When a sequence is used as a value in an expression, it renders as the
        call to C{nextval} that will produce its next value.
        """
        self.assertEquals(
            Insert({self.schema.BOZ.QUX: self.schema.A_SEQ}).toSQL(),
            SQLFragment("insert into BOZ (QUX) values (nextval('A_SEQ'))", [])
        )


    def test_nextSequenceValueOracle(self):
        """
        When a sequence is used as a value in an expression in the Oracle
        dialect, it renders as the C{nextval} attribute of the appropriate
        sequence.
        """
        self.assertEquals(
            Insert(
                {self.schema.BOZ.QUX: self.schema.A_SEQ}
            ).toSQL(QueryGenerator(ORACLE_DIALECT, FixedPlaceholder("?"))),
            SQLFragment("insert into BOZ (QUX) values (A_SEQ.nextval)", [])
        )


    def test_nextSequenceDefaultImplicitExplicitOracle(self):
        """
        In Oracle's dialect, sequence defaults can't be implemented without
        using triggers, so instead we just explicitly always include the
        sequence default value.
        """
        addSQLToSchema(
            schema=self.schema.model,
            schemaData=(
                "create table DFLTR (a varchar(255), "
                "b integer default nextval('A_SEQ'));"
            )
        )
        self.assertEquals(
            Insert({self.schema.DFLTR.a: "hello"}).toSQL(
                QueryGenerator(ORACLE_DIALECT, FixedPlaceholder("?"))
            ),
            SQLFragment("insert into DFLTR (a, b) values "
                        "(?, A_SEQ.nextval)", ["hello"]),
        )
        # Should be the same if it's explicitly specified.
        self.assertEquals(
            Insert(
                {
                    self.schema.DFLTR.a: "hello",
                    self.schema.DFLTR.b: self.schema.A_SEQ
                }
            ).toSQL(
                QueryGenerator(ORACLE_DIALECT, FixedPlaceholder("?"))
            ),
            SQLFragment(
                "insert into DFLTR (a, b) values (?, A_SEQ.nextval)", ["hello"]
            ),
        )


    def test_numericParams(self):
        """
        An L{IAsyncTransaction} with the C{numeric} paramstyle attribute will
        cause statements to be generated with parameters in the style of
        C{:1 :2 :3}, as per the DB-API.
        """
        stmts = []

        class FakeOracleTxn(object):
            def execSQL(self, text, params, exc):
                stmts.append((text, params))
            dialect = ORACLE_DIALECT
            paramstyle = "numeric"

        Select(
            [self.schema.FOO.BAR],
            From=self.schema.FOO,
            Where=(self.schema.FOO.BAR == 7).And(self.schema.FOO.BAZ == 9)
        ).on(FakeOracleTxn())

        self.assertEquals(
            stmts,
            [("select BAR from FOO where BAR = :1 and BAZ = :2", [7, 9])]
        )


    def test_rewriteOracleNULLs_Select(self):
        """
        Oracle databases cannot distinguish between the empty string and
        C{NULL}.  When you insert an empty string, C{cx_Oracle} therefore
        treats it as a C{None} and will return that when you select it back
        again.  We address this in the schema by dropping C{not null}
        constraints.

        Therefore, when executing a statement which includes a string column,
        C{on} should rewrite None return values from C{cx_Oracle} to be empty
        bytestrings, but only for string columns.
        """
        rows = resultOf(
            Select(
                [self.schema.NULLCHECK.ASTRING, self.schema.NULLCHECK.ANUMBER],
                From=self.schema.NULLCHECK
            ).on(NullTestingOracleTxn())
        )[0]

        self.assertEquals(rows, [["", None]])


    def test_rewriteOracleNULLs_SelectAllColumns(self):
        """
        Same as L{test_rewriteOracleNULLs_Select}, but with the L{ALL_COLUMNS}
        shortcut.
        """
        rows = resultOf(
            Select(From=self.schema.NULLCHECK).on(NullTestingOracleTxn())
        )[0]
        self.assertEquals(rows, [["", None]])


    def test_nestedLogicalExpressions(self):
        """
        Make sure that logical operator precedence inserts proper parenthesis
        when needed.  e.g. C{a.And(b.Or(c))} needs to be C{a and (b or c)} not
        C{a and b or c}.
        """
        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(
                    (
                        self.schema.FOO.BAR != 7
                    ).And(
                        self.schema.FOO.BAZ != 8
                    ).And(
                        (self.schema.FOO.BAR == 8).Or(self.schema.FOO.BAZ == 0)
                    )
                )
            ).toSQL(),
            SQLFragment(
                "select * from FOO where BAR != ? and BAZ != ? and "
                "(BAR = ? or BAZ = ?)",
                [7, 8, 8, 0]
            )
        )

        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(
                    (
                        self.schema.FOO.BAR != 7
                    ).Or(
                        self.schema.FOO.BAZ != 8
                    ).Or(
                        (
                            self.schema.FOO.BAR == 8
                        ).And(
                            self.schema.FOO.BAZ == 0
                        )
                    )
                )
            ).toSQL(),
            SQLFragment(
                "select * from FOO where BAR != ? or BAZ != ? or "
                "BAR = ? and BAZ = ?",
                [7, 8, 8, 0]
            )
        )

        self.assertEquals(
            Select(
                From=self.schema.FOO,
                Where=(
                    (
                        self.schema.FOO.BAR != 7
                    ).Or(
                        self.schema.FOO.BAZ != 8
                    ).And(
                        (self.schema.FOO.BAR == 8).Or(self.schema.FOO.BAZ == 0)
                    )
                )
            ).toSQL(),
            SQLFragment(
                "select * from FOO where (BAR != ? or BAZ != ?) and "
                "(BAR = ? or BAZ = ?)",
                [7, 8, 8, 0]
            )
        )


    def test_updateWithNULL(self):
        """
        As per the DB-API specification, "SQL NULL values are represented by
        the Python None singleton on input and output."  When a C{None} is
        provided as a value to an L{Update}, it will be relayed to the database
        as a parameter.
        """
        self.assertEquals(
            Update(
                {self.schema.BOZ.QUX: None},
                Where=self.schema.BOZ.QUX == 7
            ).toSQL(),
            SQLFragment("update BOZ set QUX = ? where QUX = ?", [None, 7])
        )


    def test_subSelectComparison(self):
        """
        A comparison of a column to a sub-select in a where clause will result
        in a parenthetical C{where} clause.
        """
        self.assertEquals(
            Update(
                {self.schema.BOZ.QUX: 9},
                Where=(
                    self.schema.BOZ.QUX ==
                    Select(
                        [self.schema.FOO.BAR],
                        From=self.schema.FOO,
                        Where=self.schema.FOO.BAZ == 12
                    )
                )
            ).toSQL(),
            SQLFragment(
                # NOTE: it's very important that the comparison _always_ go in
                # this order (column from the UPDATE first, inner SELECT
                # second) as the other order will be considered a syntax error.
                "update BOZ set QUX = ? "
                "where QUX = (select BAR from FOO where BAZ = ?)",
                [9, 12]
            )
        )


    def test_tupleComparison(self):
        """
        A L{Tuple} allows for simultaneous comparison of multiple values in a
        C{Where} clause.  This feature is particularly useful when issuing an
        L{Update} or L{Delete}, where the comparison is with values from a
        subselect.  (A L{Tuple} will be automatically generated upon comparison
        to a C{tuple} or C{list}.)
        """
        self.assertEquals(
            Update(
                {self.schema.BOZ.QUX: 1},
                Where=(
                    (self.schema.BOZ.QUX, self.schema.BOZ.QUUX) ==
                    Select(
                        [self.schema.FOO.BAR, self.schema.FOO.BAZ],
                        From=self.schema.FOO,
                        Where=self.schema.FOO.BAZ == 2
                    )
                )
            ).toSQL(),
            SQLFragment(
                # NOTE: it's very important that the comparison _always_ go in
                # this order (tuple of columns from the UPDATE first, inner
                # SELECT second) as the other order will be considered a syntax
                # error.
                "update BOZ set QUX = ? where (QUX, QUUX) = ("
                "select BAR, BAZ from FOO where BAZ = ?)", [1, 2]
            )
        )


    def test_tupleOfConstantsComparison(self):
        """
        For some reason Oracle requires multiple parentheses for comparisons.
        """
        self.assertEquals(
            Select(
                [self.schema.FOO.BAR],
                From=self.schema.FOO,
                Where=(
                    Tuple([self.schema.FOO.BAR, self.schema.FOO.BAZ]) ==
                    Tuple([Constant(7), Constant(9)])
                )
            ).toSQL(),
            SQLFragment(
                "select BAR from FOO where (BAR, BAZ) = ((?, ?))", [7, 9]
            )
        )


    def test_oracleTableTruncation(self):
        """
        L{Table}'s SQL generation logic will truncate table names if the
        dialect (i.e. Oracle) demands it.
        (See txdav.common.datastore.sql_tables for the schema translator and
        enforcement of name uniqueness in the derived schema.)
        """
        addSQLToSchema(
            self.schema.model,
            "create table veryveryveryveryveryveryveryverylong "
            "(foo integer);"
        )
        vvl = self.schema.veryveryveryveryveryveryveryverylong
        self.assertEquals(
            Insert({vvl.foo: 1}).toSQL(
                QueryGenerator(ORACLE_DIALECT, FixedPlaceholder("?"))
            ),
            SQLFragment(
                "insert into veryveryveryveryveryveryveryve (foo) values (?)",
                [1]
            )
        )


    def test_columnEqualityTruth(self):
        """
        Mostly in support of L{test_columnsAsDictKeys}, the "same" column
        should compare C{True} to itself and C{False} to other values.
        """
        s = self.schema
        self.assertEquals(bool(s.FOO.BAR == s.FOO.BAR), True)
        self.assertEquals(bool(s.FOO.BAR != s.FOO.BAR), False)
        self.assertEquals(bool(s.FOO.BAZ != s.FOO.BAR), True)


    def test_columnsAsDictKeys(self):
        """
        An odd corner of the syntactic sugar provided by the DAL is that the
        column objects have to participate both in augmented equality
        comparison (C{==} returns an expression object) as well as dictionary
        keys (for Insert and Update statement objects).  Therefore it should be
        possible to I{manipulate} dictionaries of keys as well.
        """
        values = {self.schema.FOO.BAR: 1}
        self.assertEquals(values, {self.schema.FOO.BAR: 1})
        values.pop(self.schema.FOO.BAR)
        self.assertEquals(values, {})



class OracleConnectionMethods(object):
    def test_rewriteOracleNULLs_Insert(self):
        """
        The behavior described in L{test_rewriteOracleNULLs_Select} applies to
        other statement types as well, specifically those with C{returning}
        clauses.
        """
        # Add 2 cursor variable values so that these will be used by
        # FakeVariable.getvalue.
        self.factory.varvals.extend([None, None])
        rows = self.resultOf(
            Insert(
                {
                    self.schema.NULLCHECK.ASTRING: "",
                    self.schema.NULLCHECK.ANUMBER: None,
                },
                Return=[
                    self.schema.NULLCHECK.ASTRING,
                    self.schema.NULLCHECK.ANUMBER,
                ]
            ).on(self.createTransaction())
        )[0]
        self.assertEquals(rows, [["", None]])


    def test_insertMultiReturnOnOracleTxn(self):
        """
        As described in L{test_insertMultiReturnOracle}, Oracle deals with
        C{returning} clauses by using out parameters.  However, this is not
        quite enough, as the code needs to actually retrieve the values from
        the out parameters.
        """
        i = Insert(
            {self.schema.FOO.BAR: 40, self.schema.FOO.BAZ: 50},
            Return=(self.schema.FOO.BAR, self.schema.FOO.BAZ)
        )
        self.factory.varvals.extend(["first val!", "second val!"])
        result = self.resultOf(i.on(self.createTransaction()))
        self.assertEquals(result, [[["first val!", "second val!"]]])
        curvars = self.factory.connections[0].cursors[0].variables
        self.assertEquals(len(curvars), 2)
        self.assertEquals(curvars[0].type, FakeCXOracleModule.NUMBER)
        self.assertEquals(curvars[1].type, FakeCXOracleModule.STRING)


    def test_insertNoReturnOracle(self):
        """
        In addition to being able to execute insert statements with a Return
        attribute, oracle also ought to be able to execute insert statements
        with no Return at all.
        """
        # This statement should return nothing from .fetchall(), so...
        self.factory.hasResults = False
        i = Insert(
            {self.schema.FOO.BAR: 40, self.schema.FOO.BAZ: 50}
        )
        result = self.resultOf(i.on(self.createTransaction()))
        self.assertEquals(result, [None])



class OracleConnectionTests(
    ConnectionPoolHelper, ExampleSchemaHelper, OracleConnectionMethods,
    TestCase
):
    """
    Tests which use an oracle connection.
    """

    dialect = ORACLE_DIALECT

    def setUp(self):
        """
        Create a fake oracle-ish connection pool without using real threads or
        a real database.
        """
        self.patch(syntax, "cx_Oracle", FakeCXOracleModule)
        super(OracleConnectionTests, self).setUp()
        ExampleSchemaHelper.setUp(self)



class OracleNetConnectionTests(
    NetworkedPoolHelper, ExampleSchemaHelper, OracleConnectionMethods,
    TestCase
):

    dialect = ORACLE_DIALECT

    def setUp(self):
        self.patch(syntax, "cx_Oracle", FakeCXOracleModule)
        super(OracleNetConnectionTests, self).setUp()
        ExampleSchemaHelper.setUp(self)
        self.pump.client.dialect = ORACLE_DIALECT



# -*- test-case-name: twext.enterprise.dal.test.test_record -*-
##
# Copyright (c) 2012-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
RECORD: Relational Entity Creation from Objects Representing Data.

This is an asynchronous object-relational mapper based on
L{twext.enterprise.dal.syntax}.
"""

__all__ = [
    "ReadOnly",
    "NoSuchRecord",
    "fromTable",
    "Record",
]

from twisted.internet.defer import inlineCallbacks, returnValue
from twext.enterprise.dal.syntax import (
    Select, Tuple, Constant, ColumnSyntax, Insert, Update, Delete, SavepointAction
)
from twext.enterprise.util import parseSQLTimestamp
# from twext.enterprise.dal.syntax import ExpressionSyntax



class ReadOnly(AttributeError):
    """
    A caller attempted to set an attribute on a database-backed record, rather
    than updating it through L{Record.update}.
    """

    def __init__(self, className, attributeName):
        self.className = className
        self.attributeName = attributeName
        super(ReadOnly, self).__init__(
            "SQL-backed attribute '{0}.{1}' is read-only. "
            "Use '.update(...)' to modify attributes."
            .format(className, attributeName)
        )



class NoSuchRecord(Exception):
    """
    No matching record could be found.
    """



class _RecordMeta(type):
    """
    Metaclass for associating a L{fromTable} with a L{Record} at inheritance
    time.
    """

    def __new__(cls, name, bases, ns):
        """
        Create a new instance of this meta-type.
        """
        newbases = []
        table = None
        namer = None

        for base in bases:
            if isinstance(base, fromTable):
                if table is not None:
                    raise RuntimeError(
                        "Can't define a class from two or more tables at once."
                    )
                table = base.table
            elif getattr(base, "table", None) is not None:
                raise RuntimeError(
                    "Can't define a record class by inheriting one already "
                    "mapped to a table."
                    # TODO: more info
                )
            else:
                if namer is None:
                    if isinstance(base, _RecordMeta):
                        namer = base
                newbases.append(base)

        if table is not None:
            attrmap = {}
            colmap = {}
            allColumns = list(table)
            for column in allColumns:
                attrname = namer.namingConvention(column.model.name)
                attrmap[attrname] = column
                colmap[column] = attrname
            ns.update(table=table, __attrmap__=attrmap, __colmap__=colmap)
            ns.update(attrmap)

        return super(_RecordMeta, cls).__new__(cls, name, tuple(newbases), ns)



class fromTable(object):
    """
    Inherit from this after L{Record} to specify which table your L{Record}
    subclass is mapped to.
    """

    def __init__(self, aTable):
        """
        @param table: The table to map to.
        @type table: L{twext.enterprise.dal.syntax.TableSyntax}
        """
        self.table = aTable



class Record(object):
    """
    Superclass for all database-backed record classes.  (i.e.  an object mapped
    from a database record).

    @cvar table: the table that represents this L{Record} in the database.
    @type table: L{TableSyntax}

    @ivar transaction: The L{IAsyncTransaction} where this record is being
        loaded.  This may be C{None} if this L{Record} is not participating in
        a transaction, which may be true if it was instantiated but never
        saved.

    @cvar __colmap__: map of L{ColumnSyntax} objects to attribute names.
    @type __colmap__: L{dict}

    @cvar __attrmap__: map of attribute names to L{ColumnSyntax} objects.
    @type __attrmap__: L{dict}
    """

    __metaclass__ = _RecordMeta

    transaction = None

    def __setattr__(self, name, value):
        """
        Once the transaction is initialized, this object is immutable.  If you
        want to change it, use L{Record.update}.
        """
        if self.transaction is not None:
            raise ReadOnly(self.__class__.__name__, name)

        return super(Record, self).__setattr__(name, value)


    def __repr__(self):
        r = (
            "<{0} record from table {1}"
            .format(self.__class__.__name__, self.table.model.name)
        )
        for k in sorted(self.__attrmap__.keys()):
            r += " {0}={1}".format(k, repr(getattr(self, k)))
        r += ">"
        return r


    def __hash__(self):
        return hash(tuple([getattr(self, attr) for attr in self.__attrmap__.keys()]))


    def __eq__(self, other):
        if type(self) != type(other):
            return False
        attrs = dict([(attr, getattr(self, attr),) for attr in self.__attrmap__.keys()])
        otherattrs = dict([(attr, getattr(other, attr),) for attr in other.__attrmap__.keys()])
        return attrs == otherattrs


    @classmethod
    def fromTable(cls, table):
        """
        Initialize from a L{Table} at run time.

        @param table: table containing the record data
        @type table: L{Table}
        """
        cls.__attrmap__ = {}
        cls.__colmap__ = {}
        allColumns = list(table)
        for column in allColumns:
            attrname = cls.namingConvention(column.model.name)
            cls.__attrmap__[attrname] = column
            cls.__colmap__[column] = attrname


    @staticmethod
    def namingConvention(columnName):
        """
        Implement the convention for naming-conversion between column names
        (typically, upper-case database names map to lower-case attribute
        names).
        """
        words = columnName.lower().split("_")

        def cap(word):
            if word.lower() in ("id", "uid", "guid",):
                return word.upper()
            else:
                return word.capitalize()

        return words[0] + "".join(map(cap, words[1:]))


    @classmethod
    def _primaryKeyExpression(cls):
        return Tuple([ColumnSyntax(c) for c in cls.table.model.primaryKey])


    def _primaryKeyValue(self):
        val = []
        for col in self._primaryKeyExpression().columns:
            val.append(getattr(self, self.__class__.__colmap__[col]))
        return val


    @classmethod
    def _primaryKeyComparison(cls, primaryKey):
        return cls._primaryKeyExpression() == Tuple(map(Constant, primaryKey))


    @classmethod
    @inlineCallbacks
    def load(cls, transaction, *primaryKey):
        results = yield cls.query(
            transaction,
            cls._primaryKeyComparison(primaryKey)
        )
        if len(results) != 1:
            raise NoSuchRecord()
        else:
            returnValue(results[0])


    @classmethod
    @inlineCallbacks
    def create(cls, transaction, **k):
        """
        Create a row.

        Used like this::

            MyRecord.create(transaction, column1=1, column2=u"two")
        """
        self = cls.make(**k)
        yield self.insert(transaction)
        returnValue(self)


    @classmethod
    def make(cls, **k):
        """
        Make a record without creating one in the database - this will not have an
        associated L{transaction}. When the record is ready to be written to the database
        use L{SerializeableRecord.insert} to add it. Before it gets written to the DB, the
        attributes can be changed.
        """
        self = cls()
        attrtocol = cls.__attrmap__

        for attr in attrtocol:
            col = attrtocol[attr]
            if attr in k:
                value = k.pop(attr)
                setattr(self, attr, value)
            else:
                if col.model.needsValue():
                    raise TypeError(
                        "required attribute {0!r} not passed"
                        .format(attr)
                    )

        if k:
            raise TypeError("received unknown attribute{0}: {1}".format(
                "s" if len(k) > 1 else "", ", ".join(sorted(k))
            ))

        return self


    def duplicate(self):
        return self.make(**dict([(attr, getattr(self, attr),) for attr in self.__attrmap__.keys()]))


    def isnew(self):
        return self.transaction is None


    def _attributesFromRow(self, attributeList):
        """
        Take some data loaded from a row and apply it to this instance,
        converting types as necessary.

        @param attributeList: a C{list} of 2-C{tuples} of C{(attributeName,
            attributeValue)}.
        """
        for setAttribute, setValue in attributeList:
            setColumn = self.__attrmap__[setAttribute]
            if setColumn.model.type.name == "timestamp" and setValue is not None:
                setValue = parseSQLTimestamp(setValue)
            setattr(self, setAttribute, setValue)


    @inlineCallbacks
    def insert(self, transaction):
        """
        Insert a new a row for an existing record that was not initially created in the database.
        """

        # Cannot do this if a transaction has already been assigned because that means
        # the record already exists in the DB.

        if self.transaction is not None:
            raise ReadOnly(self.__class__.__name__, "Cannot insert")

        colmap = {}
        attrtocol = self.__attrmap__
        needsCols = []
        needsAttrs = []

        for attr in attrtocol:
            col = attrtocol[attr]
            v = getattr(self, attr)
            if not isinstance(v, ColumnSyntax):
                colmap[col] = v
            else:
                if col.model.needsValue():
                    raise TypeError(
                        "required attribute {0!r} not passed"
                        .format(attr)
                    )
                else:
                    needsCols.append(col)
                    needsAttrs.append(attr)

        result = yield (Insert(colmap, Return=needsCols if needsCols else None)
                        .on(transaction))
        if needsCols:
            self._attributesFromRow(zip(needsAttrs, result[0]))

        self.transaction = transaction


    def delete(self):
        """
        Delete this row from the database.

        @return: a L{Deferred} which fires with C{None} when the underlying row
            has been deleted, or fails with L{NoSuchRecord} if the underlying
            row was already deleted.
        """
        return Delete(
            From=self.table,
            Where=self._primaryKeyComparison(self._primaryKeyValue())
        ).on(self.transaction, raiseOnZeroRowCount=NoSuchRecord)


    @inlineCallbacks
    def update(self, **kw):
        """
        Modify the given attributes in the database.

        @return: a L{Deferred} that fires when the updates have been sent to
            the database.
        """
        colmap = {}
        for k, v in kw.iteritems():
            colmap[self.__attrmap__[k]] = v

        yield Update(
            colmap,
            Where=self._primaryKeyComparison(self._primaryKeyValue())
        ).on(self.transaction)

        self.__dict__.update(kw)


    @inlineCallbacks
    def lock(self, where=None):
        """
        Lock with a select for update.

        @param where: SQL expression used to match the rows to lock, by default this is just an expression
            that matches the primary key of this L{Record}, but it can be used to lock multiple L{Records}
            matching the expression in one go. If it is an L{str}, then all rows will be matched.
        @type where: L{SQLExpression} or L{None}
        @return: a L{Deferred} that fires when the lock has been acquired.
        """
        if where is None:
            where = self._primaryKeyComparison(self._primaryKeyValue())
        elif isinstance(where, str):
            where = None
        yield Select(
            list(self.table),
            From=self.table,
            Where=where,
            ForUpdate=True,
        ).on(self.transaction)


    @inlineCallbacks
    def trylock(self, where=None):
        """
        Try to lock with a select for update no wait. If it fails, rollback to
        a savepoint and return L{False}, else return L{True}.

        @param where: SQL expression used to match the rows to lock, by default this is just an expression
            that matches the primary key of this L{Record}, but it can be used to lock multiple L{Records}
            matching the expression in one go. If it is an L{str}, then all rows will be matched.
        @type where: L{SQLExpression} or L{None}
        @return: a L{Deferred} that fires when the updates have been sent to
            the database.
        """

        if where is None:
            where = self._primaryKeyComparison(self._primaryKeyValue())
        elif isinstance(where, str):
            where = None
        savepoint = SavepointAction("Record_trylock_{}".format(self.__class__.__name__))
        yield savepoint.acquire(self.transaction)
        try:
            yield Select(
                list(self.table),
                From=self.table,
                Where=where,
                ForUpdate=True,
                NoWait=True,
            ).on(self.transaction)
        except:
            yield savepoint.rollback(self.transaction)
            returnValue(False)
        else:
            yield savepoint.release(self.transaction)
            returnValue(True)


    @classmethod
    def pop(cls, transaction, *primaryKey):
        """
        Atomically retrieve and remove a row from this L{Record}'s table
        with a primary key value of C{primaryKey}.

        @return: a L{Deferred} that fires with an instance of C{cls}, or fails
            with L{NoSuchRecord} if there were no records in the database.
        @rtype: L{Deferred}
        """
        return cls._rowsFromQuery(
            transaction,
            Delete(
                Where=cls._primaryKeyComparison(primaryKey),
                From=cls.table,
                Return=list(cls.table)
            ),
            lambda: NoSuchRecord()
        ).addCallback(lambda x: x[0])


    @classmethod
    def query(cls, transaction, expr, order=None, group=None, limit=None, forUpdate=False, noWait=False, ascending=True, distinct=False):
        """
        Query the table that corresponds to C{cls}, and return instances of
        C{cls} corresponding to the rows that are returned from that table.

        @param expr: An L{ExpressionSyntax} that constraints the results of the
            query.  This is most easily produced by accessing attributes on the
            class; for example, C{MyRecordType.query((MyRecordType.col1 >
            MyRecordType.col2).And(MyRecordType.col3 == 7))}

        @param order: A L{ColumnSyntax} to order the resulting record objects
            by.

        @param ascending: A boolean; if C{order} is not C{None}, whether to
            sort in ascending or descending order.

        @param group: a L{ColumnSyntax} to group the resulting record objects
            by.

        @param forUpdate: do a SELECT ... FOR UPDATE
        @type forUpdate: L{bool}
        @param noWait: include NOWAIT with the FOR UPDATE
        @type noWait: L{bool}
        """
        return cls._rowsFromQuery(
            transaction,
            cls.queryExpr(
                expr,
                order=order,
                group=group,
                limit=limit,
                forUpdate=forUpdate,
                noWait=noWait,
                ascending=ascending,
                distinct=distinct,
            ),
            None
        )


    @classmethod
    def queryExpr(cls, expr, attributes=None, order=None, group=None, limit=None, forUpdate=False, noWait=False, ascending=True, distinct=False):
        """
        Query expression that corresponds to C{cls}. Used in cases where a sub-select
        on this record's table is needed.

        @param expr: An L{ExpressionSyntax} that constraints the results of the
            query.  This is most easily produced by accessing attributes on the
            class; for example, C{MyRecordType.query((MyRecordType.col1 >
            MyRecordType.col2).And(MyRecordType.col3 == 7))}

        @param order: A L{ColumnSyntax} to order the resulting record objects
            by.

        @param ascending: A boolean; if C{order} is not C{None}, whether to
            sort in ascending or descending order.

        @param group: a L{ColumnSyntax} to group the resulting record objects
            by.

        @param forUpdate: do a SELECT ... FOR UPDATE
        @type forUpdate: L{bool}
        @param noWait: include NOWAIT with the FOR UPDATE
        @type noWait: L{bool}
        """
        kw = {}
        if order is not None:
            kw.update(OrderBy=order, Ascending=ascending)
        if group is not None:
            kw.update(GroupBy=group)
        if limit is not None:
            kw.update(Limit=limit)
        if forUpdate:
            kw.update(ForUpdate=True)
            if noWait:
                kw.update(NoWait=True)
        if distinct:
            kw.update(Distinct=True)
        if attributes is None:
            attributes = list(cls.table)
        return Select(
            attributes,
            From=cls.table,
            Where=expr,
            **kw
        )


    @classmethod
    def querysimple(cls, transaction, **kw):
        """
        Match all rows matching the specified attribute/values from the table that corresponds to C{cls}.
        All attributes are logically AND'ed.
        """
        where = None
        for k, v in kw.iteritems():
            subexpr = (cls.__attrmap__[k] == v)
            if where is None:
                where = subexpr
            else:
                where = where.And(subexpr)
        return cls.query(transaction, where)


    @classmethod
    def all(cls, transaction):
        """
        Load all rows from the table that corresponds to C{cls} and return
        instances of C{cls} corresponding to all.
        """
        return cls._rowsFromQuery(
            transaction,
            Select(
                list(cls.table),
                From=cls.table,
                OrderBy=cls._primaryKeyExpression()
            ),
            None
        )


    @classmethod
    def updatesome(cls, transaction, where, **kw):
        """
        Update rows matching the where expression from the table that corresponds to C{cls}.
        """
        colmap = {}
        for k, v in kw.iteritems():
            colmap[cls.__attrmap__[k]] = v

        return Update(
            colmap,
            Where=where
        ).on(transaction)


    @classmethod
    def deleteall(cls, transaction):
        """
        Delete all rows from the table that corresponds to C{cls}.
        """
        return cls.deletesome(transaction, None)


    @classmethod
    def deletesome(cls, transaction, where, returnCols=None):
        """
        Delete all rows matching the where expression from the table that corresponds to C{cls}.
        """
        return Delete(
            From=cls.table,
            Where=where,
            Return=returnCols,
        ).on(transaction)


    @classmethod
    def deletesimple(cls, transaction, **kw):
        """
        Delete all rows matching the specified attribute/values from the table that corresponds to C{cls}.
        All attributes are logically AND'ed.
        """
        where = None
        for k, v in kw.iteritems():
            subexpr = (cls.__attrmap__[k] == v)
            if where is None:
                where = subexpr
            else:
                where = where.And(subexpr)
        return cls.deletesome(transaction, where)


    @classmethod
    @inlineCallbacks
    def _rowsFromQuery(cls, transaction, qry, rozrc):
        """
        Execute the given query, and transform its results into instances of
        C{cls}.

        @param transaction: an L{IAsyncTransaction} to execute the query on.

        @param qry: a L{_DMLStatement} (XXX: maybe _DMLStatement or some
            interface that defines "on" should be public?) whose results are
            the list of columns in C{self.table}.

        @param rozrc: The C{raiseOnZeroRowCount} argument.

        @return: a L{Deferred} that succeeds with a C{list} of instances of
            C{cls} or fails with an exception produced by C{rozrc}.
        """
        rows = yield qry.on(transaction, raiseOnZeroRowCount=rozrc)
        selves = []
        names = [cls.__colmap__[column] for column in list(cls.table)]
        for row in rows:
            self = cls()
            self._attributesFromRow(zip(names, row))
            self.transaction = transaction
            selves.append(self)
        returnValue(selves)



class SerializableRecord(Record):
    """
    An L{Record} that serializes/deserializes its attributes for a text-based
    transport (e.g., JSON-over-HTTP) to allow records to be transferred from
    one system to another (with potentially mismatched schemas).
    """

    def serialize(self):
        """
        Create an L{dict} of each attribute with L{str} values for each attribute
        value. Sub-classes may need to override this to specialize certain value
        conversions.

        @return: mapping of attribute to string values
        @rtype: L{dict} of L{str}:L{str}
        """

        # Certain values have to be mapped to str
        result = {}
        for attr in self.__attrmap__:
            value = getattr(self, attr)
            col = self.__attrmap__[attr]
            if col.model.type.name == "timestamp" and value is not None:
                value = str(value)
            result[attr] = value
        return result


    @classmethod
    def deserialize(cls, attrmap):
        """
        Given an L{dict} mapping attributes to values, create an L{Record} with
        the specified values. Sub-classes may need to override this to handle special
        values that need to be converted to specific types. They also need to override
        this to handle possible schema mismatches (attributes no longer used, new
        attributes not present in the map).

        @param attrmap: serialized representation of a record
        @type attrmap: L{dict} of L{str}:L{str}

        @return: a newly created, but not inserted, record
        @rtype: L{SerializableRecord}
        """

        # Certain values have to be mapped to non-str types
        mapped = {}
        for attr in attrmap:
            value = attrmap[attr]
            col = cls.__attrmap__[attr]
            if col.model.type.name == "timestamp" and value is not None:
                value = parseSQLTimestamp(value)
            mapped[attr] = value

        record = cls.make(**mapped)
        return record



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
OpenDirectory directory service.
"""

__all__ = [
    "OpenDirectoryError",
    "OpenDirectoryConnectionError",
    "OpenDirectoryQueryError",
    "OpenDirectoryDataError",
    "DirectoryService",
    "NoQOPDigestCredentialFactory",
    "RecordType",
]


from ._service import (
    OpenDirectoryError, OpenDirectoryConnectionError, OpenDirectoryQueryError,
    OpenDirectoryDataError,
    DirectoryService,
    NoQOPDigestCredentialFactory,
)
